import os
import zipfile
import logging
import shutil
import gzip
from pathlib import Path
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from .read_data_files import read_data_files

from .config import CONFIG


def load(
    trange=["2012-11-05/00:00:00", "2012-11-06/00:00:00"],
    resolution=10,
    dtype=None,
    no_download=False,
    downloadonly=False,
    out_type="np",
    save_pickle=False,
    spdf=False,
    force_download=False,
):
    """
    This function downloads and reads SECS/EICS data.

    It does not return tplot variables, but rather the data itself.
    Usually, this data is used by makeplots.py to create contour plots.

    Parameters
    ----------
    trange : list of str
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD', 'YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss', 'YYYY-MM-DD/hh:mm:ss'].
        Default: ["2012-11-05/00:00:00", "2012-11-06/00:00:00"]
    resolution : int, optional
        Time resolution of the data in seconds.
        Default: 10
    dtype : str, optional
        Data type; Valid options: 'EICS', 'SECS'.
        Default: None
    no_download : bool, optional
        Only load data from the local cache.
        Default: False
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load data from them.
        If True, the function returns a list of the downloaded files.
        Default: False
    out_type : str, optional
        The return type: 'np' for numpy array, 'df' for pandas dataframe, 'dc' for dictionary.
        Default: 'np'
    save_pickle : bool, optional
        Whether to save the output as a pickle file.
        Default: False
    spdf : bool, optional
        If True, download data from SPDF instead of UCLA.
        Default: False
    force_download : bool, optional
        If True, re-download the data files.
        Default: False

    Returns
    -------
    Varies (list of str, np.ndarray, pd.DataFrame, or dict)

        If downloadonly is True, the function returns a list of the downloaded files.
        Otherwise, it returns the data read from the files in the specified format.
        If out_type is 'np', the data is a numpy array.
        If out_type is 'df', the data is a pandas dataframe.
        If out_type is 'dc', the data is a dictionary.

    Examples
    --------
    >>> import pyspedas
    >>> secs_vars = pyspedas.projects.secs.data(dtype='EICS', trange=['2018-02-01', '2018-02-02'], downloadonly=True)
    >>> print(secs_vars)
    >>> ['/Users/user/data/secs/EICS/2018/02/EICS20180201.zip']
    """

    if dtype is None:
        logging.error("No data type provided.")
        return None
    else:
        dtype = dtype.upper()

    if trange is None or len(trange) != 2 or trange[1] < trange[0]:
        logging.error("Invalid time range provided.")
        return None

    if not downloadonly:
        if out_type is None:
            logging.error("No out_type provided.")
            return None
        else:
            out_type = out_type.lower()
        if out_type not in ["np", "df", "dc"]:
            logging.error("Invalid out_type provided.")
            return None

    if dtype == "EICS" or dtype == "SECS":

        pathformat_prefix = dtype + "/%Y/%m/"
        pathformat_zip = pathformat_prefix + dtype + "%Y%m%d.zip"
        pathformat_gz = pathformat_prefix + dtype + "%Y%m%d.zip.gz"  # only 2007!
        pathformat_unzipped = pathformat_prefix + "%d/" + dtype + "%Y%m%d_%H%M%S.dat"
        remote_path = CONFIG["remote_data_dir"]

    else:
        logging.error("Invalid data type provided:" + str(dtype))
        return None

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat_zip, trange=trange)
    remote_names_gz = dailynames(file_format=pathformat_gz, trange=trange)
    remote_names_gz = [s for s in remote_names_gz if s[-15:-11] == "2007"]

    out_files = []
    out_files_zip = []

    if spdf:
        # Download files from SPDF
        remote_path_spdf = CONFIG["remote_data_dir_spdf"]
        # Paths at SPDF do not include the month
        # zip files at UCLA are inside a directory /year/month/
        # but the same files at SPDF are inside a directory /year/
        pathformat_prefix_spdf = dtype + "/%Y/"
        pathformat_zip_spdf = pathformat_prefix_spdf + dtype + "%Y%m%d.zip"
        remote_names_spdf = dailynames(file_format=pathformat_zip_spdf, trange=trange)
        local_dir_spdf = CONFIG["local_data_dir"]

        files_zip_spdf = download(
            remote_file=remote_names_spdf,
            remote_path=remote_path_spdf,
            local_path=local_dir_spdf,
            no_download=no_download,
            force_download=force_download,
        )
        # At this point, the zip files are downloaded but in a different dir than UCLA files.
        # We need to move them to a directory containing the month
        files_zip = []
        for f in files_zip_spdf:
            if os.path.exists(f):
                stem = Path(f).stem
                if len(stem) >= 10:
                    # Stem is similar to 'SECS20120301' or 'EICS20120301'
                    month = stem[8:10]
                    path = str(Path(f).parent)
                    new_path = os.path.join(path, month)
                    new_filename = os.path.join(new_path, os.path.basename(f))
                    if os.path.exists(new_filename):
                        os.remove(new_filename)
                    else:
                        os.makedirs(new_path)
                    new_filename = shutil.copy(f, new_path)
                    files_zip.append(new_filename)

        # SPDF does not contain any .gz files
        files_gz = []
    else:
        # Download files from UCLA
        files_zip = download(
            remote_file=remote_names,
            remote_path=remote_path,
            local_path=CONFIG["local_data_dir"],
            no_download=no_download,
            force_download=force_download,
        )
        files_gz = download(
            remote_file=remote_names_gz,
            remote_path=remote_path,
            local_path=CONFIG["local_data_dir"],
            no_download=no_download,
            force_download=force_download,
        )

    files_zip = files_zip + files_gz

    if files_zip is not None:
        for rf_zip_zero in files_zip:
            if rf_zip_zero.endswith(".gz"):
                rf_zip = rf_zip_zero[0:-3]
                # unzip .gz file to .zip file
                with gzip.open(rf_zip_zero, "rb") as f_in:
                    with open(rf_zip, "wb") as f_out:
                        shutil.copyfileobj(f_in, f_out)
            elif rf_zip_zero.endswith(".zip"):
                rf_zip = rf_zip_zero
            else:
                rf_zip = rf_zip_zero
            out_files_zip.append(rf_zip)
            # print('Start for unzipping process ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
            foldername_unzipped = rf_zip[0:-19] + rf_zip[-8:-6] + "/" + rf_zip[-6:-4]

            # print('foldername_unzipped-------: ', foldername_unzipped)
            ### add??????
            if not os.path.isdir(foldername_unzipped):
                logging.info("Start unzipping: " + rf_zip + "  ------")
                with zipfile.ZipFile(rf_zip, "r") as zip_ref:
                    zip_ref.extractall(rf_zip[0:-16])
                if not os.path.isdir(foldername_unzipped):
                    # for the case of unzipping directly without the %d folder made.
                    # make %d folder
                    os.makedirs(foldername_unzipped)
                    # move .dat files
                    sourcepath = rf_zip[0:-16]
                    sourcefiles = os.listdir(sourcepath)
                    destinationpath = foldername_unzipped
                    logging.info("start to move files: --------------")
                    for file in sourcefiles:
                        if rf_zip[-16:-4] in file and file.endswith(".dat"):
                            shutil.move(
                                os.path.join(sourcepath, file),
                                os.path.join(destinationpath, file),
                            )

            else:
                logging.info(
                    "Unzipped folder: "
                    + foldername_unzipped
                    + " existed, skip unzipping  ------"
                )

    if files_zip is not None:
        for file in files_zip:
            out_files.append(file)
    out_files = sorted(out_files)

    if out_files_zip is not None:
        out_files_zip = list(set(out_files_zip))
        out_files_zip = sorted(out_files_zip)

    if downloadonly:
        return out_files_zip  # out_files

    remote_names_unzipped = dailynames(
        file_format=pathformat_unzipped, trange=trange, res=resolution
    )
    """
    files_unzipped = download(remote_file=remote_names_unzipped, remote_path=CONFIG['remote_data_dir'],
                         local_path=CONFIG['local_data_dir'], no_download=True)
    """
    remote_names_unzipped_existed = [
        rnud
        for rnud in remote_names_unzipped
        for ofz in out_files_zip
        if ofz[-16:-4] in rnud
    ]
    remote_names_unzipped = remote_names_unzipped_existed
    out_files_unzipped = [
        os.path.join(CONFIG["local_data_dir"], rf_res)
        for rf_res in remote_names_unzipped
    ]
    out_files_unzipped = sorted(out_files_unzipped)

    if out_files_unzipped == []:
        data_vars = []
    else:
        data_vars = read_data_files(
            out_files=out_files_unzipped,
            dtype=dtype,
            out_type=out_type,
            save_pickle=save_pickle,
        )
        # print('data_vars: ', data_vars, np.shape(data_vars))

    return data_vars



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas

def tplot_names(quiet=False):
    """
    This function will print out and return a list of all current Tplot Variables stored in the memory.  
    
    Parameters
    ----------
        quiet : bool
            If True, does not print out the variables (only returns the list variables)
         
    Returns
    -------
        list of str
            A list of all Tplot Variables stored in the memory
            
    Examples
    --------
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> tnames = pyspedas.tplot_names()
        0 : Variable 1

    """
    
    index = 0
    return_names=[]

    # TODO: Print out links as well?

    for key, _ in pyspedas.tplot_tools.data_quants.items():
        if isinstance(pyspedas.tplot_tools.data_quants[key], dict):
            # non-record varying variables are stored as dictionaries
            if isinstance(key, str):
                names_to_print = key

            if quiet != True:
                print(index, ":", names_to_print)

            return_names.append(names_to_print)
            index += 1
            continue

        if len(pyspedas.tplot_tools.data_quants[key].attrs['plot_options']['overplots_mpl']) != 0:
            if quiet:
                # In this context we only want variable names, and no other formatting
                names_to_print = pyspedas.tplot_tools.data_quants[key].name
            else:
                names_to_print = pyspedas.tplot_tools.data_quants[key].name + "  data from: "
                for oplot_name in pyspedas.tplot_tools.data_quants[key].attrs['plot_options']['overplots_mpl']:
                    names_to_print = names_to_print + " " + oplot_name
        else:
            if isinstance(key, str):
                names_to_print = pyspedas.tplot_tools.data_quants[key].name

        if quiet != True:
            print(index, ":", names_to_print)
            
        index += 1

        return_names.append(names_to_print)
    return return_names


"""Parameters for pyspedas testing files.

Specify local output directory for testing, and remote directory with SPEDAS validation files.
"""

import os

TESTING_CONFIG = {
    "local_testing_dir": "_testing_output",  # Testing output will be saved here
    "remote_validation_dir": "https://github.com/spedas/test_data/raw/refs/heads/main/",  # SPEDAS-pyspedas validation files can be found here
    "global_display": False,  # Whether to display plots during testing
}

# Override local output directory with environment variables, if there are any
if os.environ.get("SPEDAS_DATA_DIR"):
    TESTING_CONFIG["local_testing_dir"] = os.sep.join(
        [os.environ["SPEDAS_DATA_DIR"], "_testing_output"]
    )
# Override local output directory with a local path
# TESTING_CONFIG["local_testing_dir"] = os.path.expanduser("~/data/_testing_output")

if os.environ.get("PYSPEDAS_TESTING_DIR"):
    TESTING_CONFIG["local_testing_dir"] = os.environ["PYSPEDAS_TESTING_DIR"]

# Override remote validation directory with environment variables, if there are any
if os.environ.get("PYSPEDAS_VALIDATION_DIR"):
    TESTING_CONFIG["remote_validation_dir"] = os.environ["PYSPEDAS_VALIDATION_DIR"]

# Override remote validation directory with a local path for offline testing
# TESTING_CONFIG["remote_validation_dir"] = os.path.expanduser("~/work/GitHub/test_data")

# Override global display setting with environment variables, if there are any
if os.environ.get("PYSPEDAS_GLOBAL_DISPLAY"):
    TESTING_CONFIG["global_display"] = bool(os.environ["PYSPEDAS_GLOBAL_DISPLAY"])

# Override global display setting for testing
# TESTING_CONFIG["global_display"] = True


def test_data_download_file(validation_dir, sub_dir, file_name, output_dir):
    """
    Fetch a validation tplot file_name from the configured remote directory or from local directory.

    This function is used in unit tests to download validation data files needed for testing.

    Parameters
    ----------
    validation_dir : str
        Base URL or path that hosts validation data files.
        For example: "https://github.com/spedas/test_data/raw/refs/heads/main/"
    sub_dir : str
        Remote subdirectory that contains the desired sub_dir/file_name.
        For example: "analysis_tools"
    file_name : str
        Name of the file to download.
        For example: "wavelet_test.tplot"
    output_dir : str
        Local destination directory for downloaded files.
        For example: "/data/_testing_output"

    Returns
    -------
    str
        Full path to the downloaded validation file, or an empty string if the
        file could not be retrieved.
    """

    import os
    import logging
    from pyspedas.utilities.download import download

    filename = ""
    remote_file = os.path.join(sub_dir, file_name)
    # Download the file if it doesn't already exist locally
    local_file = os.path.join(validation_dir, remote_file)
    if not os.path.exists(local_file):
        datafile = download(
            remote_file=remote_file,
            remote_path=validation_dir,
            local_path=output_dir,
            no_download=False,
        )
    else:
        datafile = [local_file]  # File already exists locally

    if not datafile:
        # Skip tests
        logging.info("Cannot download data validation file. Filename: " + remote_file)
        return filename

    filename = datafile[0]

    return filename



'''
Convert spherical coordinates (r, theta, phi) to Cartesian coordinates (x, y, z).

This function is similar to spc2cart.pro in IDL SPEDAS.
'''
import numpy as np


def spc2cart(r, theta, phi):
    """
    Convert spherical coordinates (r, theta, phi) to Cartesian coordinates (x, y, z).

    Parameters:
    -----------
    r : float
        Radial distance from the origin.
    theta : float
        Polar angle (angle from the positive z-axis to the point) in radians. It ranges from [0, pi].
    phi : float
        Azimuthal angle (angle in the xy-plane from the positive x-axis to the point) in radians. It ranges from [0, 2*pi].

    Returns:
    --------
    x : float
        X coordinate in Cartesian system.
    y : float
        Y coordinate in Cartesian system.
    z : float
        Z coordinate in Cartesian system.

    Notes:
    ------
    - Uses the following equations for conversion:
        x = r*sin(theta)*cos(phi)
        y = r*sin(theta)*sin(phi)
        z = r*cos(theta)
    """
    r, theta, phi = np.array(r), np.array(theta), np.array(phi)
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    return x, y, z



"""
Transform datetimes from decimal to string.
"""
from datetime import datetime, timezone
from pyspedas.tplot_tools import time_float

def time_string_one(float_time=None, fmt=None):
    """
    Transforms a single float daytime value into a string representation.

    Parameters
    ----------
    float_time : float, optional
        The input time as a float. If not provided, the current time is used. If input is actually a string,
        it is returned as-is.
    fmt : str, optional
        The format string for time conversion. The default format is '%Y-%m-%d %H:%M:%S.%f'.

    Returns
    -------
    str
        The datetime as a string formatted according to `fmt`. If `float_time` is not provided,
        returns the current datetime formatted as specified.

    Examples
    --------
    >>> from pyspedas import time_string_one
    >>> time_string_one(1679745600.0)
    '2023-03-25 12:00:00.000000'

    >>> time_string_one(1679745600.0, '%Y-%m-%d')
    '2023-03-25'

    >>> time_string_one()
    # Returns the current datetime in the default format
    """

    if fmt is None:
        fmt = '%Y-%m-%d %H:%M:%S.%f'

    if float_time is None:
        str_time = datetime.now().strftime(fmt)
    elif isinstance(float_time, str):
        # It's already a string, just return the input
        return float_time
    else:
        str_time = datetime.fromtimestamp(float_time,timezone.utc).strftime(fmt)

    return str_time


def time_string(float_time=None, fmt=None):
    """
    Transforms a list of float daytime values into a list of string representations.

    Parameters
    ----------
    float_time : floats, or list of floats, optional
        The input time(s) as float(s). If not provided, the current time is used. Any string inputs are returned as-is.
    fmt : str, optional
        The format string for time conversion. The default format is '%Y-%m-%d %H:%M:%S.%f'.

    Returns
    -------
    str or list of str
        The datetimes as strings formatted according to `fmt`. If `float_time` is not provided, returns
        the current datetime formatted as specified. If `float_time` is a single float, returns
        a single datetime string.

    Examples
    --------
    >>> time_string(1679745600.0)
    '2023-03-25 12:00:00.000000'

    >>> time_string([1679745600.0, 1679832000.0], "%Y-%m-%d %H:%M:%S")
    ['2023-03-25 12:00:00', '2023-03-26 12:00:00']

    >>> time_string()
    #'Returns the current datetime in the default format'

    Notes
    -----
    Compare to https://www.epochconverter.com/
    """
    if float_time is None:
        return time_string_one(None, fmt)
    else:
        if isinstance(float_time, (int, float)):
            return time_string_one(float_time, fmt)
        else:
            time_list = list()
            for t in float_time:
                time_list.append(time_string_one(t, fmt))
            return time_list


def time_datetime(time=None, tz=None):
    """
    Transforms a list of float daytime values or strings to a list of pythonic `datetime.datetime` objects.

    Parameters
    ----------
    time : float, list of floats, str, or list of strs, optional
        The input time(s) as float(s) representing UNIX timestamps or strings.
        If not provided, the current time is used.
    tz : datetime.timezone, optional
        The timezone for the datetime object(s).
        If not provided, UTC is used.

    Returns
    -------
    datetime.datetime or list of datetime.datetime
        The datetimes as `datetime.datetime` objects. If `time` is a single value, returns a single `datetime.datetime` object;
        if `time` is a list, returns a list of `datetime.datetime` objects. If `time` is None, returns the current datetime in UTC.

    Examples
    --------
    >>> from pyspedas import time_datetime
    >>> from datetime import timezone, timedelta
    >>> time_datetime(1679745600.0)
    # or
    >>> time_datetime('2023-03-25 12:00:00')
    # Returns a datetime object for 2023-03-25 12:00:00 UTC
    datetime.datetime(2023, 3, 25, 12, 0, tzinfo=datetime.timezone.utc)

    >>> time_datetime([1679745600.0, 1679832000.0])
    # or
    >>> time_datetime(['2023-03-25 12:00:00', '2023-03-26 12:00:00'])
    # Returns a list of datetime objects for 2023-03-25 12:00:00 UTC and 2023-03-26 12:00:00 UTC
    [datetime.datetime(2023, 3, 25, 12, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2023, 3, 26, 12, 0, tzinfo=datetime.timezone.utc)]

    >>> time_datetime()
    # Returns the current datetime object in UTC

    >>> time_datetime(1679745600.0, tz=timezone(timedelta(hours=-6)))
    # Returns a datetime object for 2023-03-25 06:00:00 in a timezone that is -6 hour from UTC
    datetime.datetime(2023, 3, 25, 6, 0, tzinfo=datetime.timezone(datetime.timedelta(days=-1, seconds=64800)))

    """

    if tz is None:
        tz = timezone.utc

    if time is None:
        return datetime.now()

    if isinstance(time, str):
        return time_datetime(time_float(time))

    if isinstance(time, (int, float)):
        return datetime.fromtimestamp(time, tz=tz)

    return [time_datetime(_time) for _time in time]



from functools import partial
from .load import load

# Alias for load function with instrument parameter set
mag = partial(load, instrument="mag")
swea = partial(load, instrument="swea")
swia = partial(load, instrument="swia")
static = partial(load, instrument="static")
sep = partial(load, instrument="sep")
kp = partial(load, instrument="kp")




import numpy as np
import pyspedas
from pyspedas.tplot_tools import get_data, is_pseudovariable
import logging

def count_traces(tvar):
    """
    Count the total number of line traces in a variable or list of variables

    Parameters
    ----------
    tvar: str or list of str
        tplot variables with traces to be counted

    Returns
    -------
    int
        Number of traces found in the input variables.  Spectrograms are not counted.

    Examples
    --------
        >>> import pyspedas
        >>> pyspedas.store_data('a', data={'x': range(10), 'y': range(10)})
        >>> pyspedas.store_data('b', data={'x': range(10), 'y': range(10)})
        >>> pyspedas.store_data('pseudovar', data=['a','b'])
        >>> pyspedas.count_traces('a')  # 1
        >>> pyspedas.count_traces('pseudovar') # 2

    """
    trace_count = 0
    if not isinstance(tvar,list):
        tvar = [tvar]
    for v in tvar:
        if v in pyspedas.tplot_tools.data_quants.keys():
            data=get_data(v, dt=True)

            if is_pseudovariable(v):
                components = pyspedas.tplot_tools.data_quants[v].attrs['plot_options']['overplots_mpl']
                trace_count += count_traces(components)
            else:
                plot_extras = pyspedas.tplot_tools.data_quants[v].attrs['plot_options']['extras']
                if plot_extras.get('spec') is not None:
                    spec = plot_extras['spec']
                else:
                    spec = False

                if len(data.y.shape) == 1:
                    num_lines = 1
                else:
                    num_lines = data.y.shape[1]

                if not spec:
                    trace_count += num_lines
        else:
            logging.warning('The name %s is not in pyspedas',v)
    return trace_count



import numpy as np
from pyspedas.tplot_tools import get_data, options, store_data, tplot_names

from .erg_interpolate_att import erg_interpolate_att
from .vector_rotate import vector_rotate


def sgi2dsi(name_in=None,
            name_out=None,
            DSI2SGI=False,
            noload=False):
    """
    This transform a time series data between the SGI and DSI coordinate systems.

    Parameters:

        name_in : str
            input tplot variable to be transformed

        name_out : str
            Name of the tplot variable in which the transformed data is stored

        DSI2SGI : bool
             Set to transform data from DSI to SGI (despun coord --> spinning coord).
             If not set, it transforms data from SGI to DSI (spinning coord --> despun coord).

    Returns:
        None

    """
    if (name_in is None) or (name_in not in tplot_names(quiet=True)):
        print('Input of Tplot name is undifiend')
        return

    if name_out is None:
        print('Tplot name for output is undifiend')
        name_out = 'result_of_sgi2dsi'

    # prepare for transformed Tplot Variable
    reload = not noload
    dl_in = get_data(name_in, metadata=True)
    get_data_array = get_data(name_in)
    time_array = get_data_array[0]
    time_length = time_array.shape[0]
    dat = get_data_array[1]

    # Get the SGA and SGI axes by interpolating the attitude data
    interpolated_values = erg_interpolate_att(name_in, noload=noload)
    sgix2ssix_angle = interpolated_values['sgiz_j2000']['y'][:, 0]
    # [deg] Now the constant angle is used, which is not correct, though
    sgix2ssix_angle[:] = 90. + 21.6

    spperiod = interpolated_values['spinperiod']['y']
    spphase = interpolated_values['spinphase']['y']
    rot_axis = np.array([[0., 0., 1.]]*time_length)

    # SGI --> DSI (despin)
    if not DSI2SGI:
        print('SGI --> DSI')
        coor_out = 'dsi'
        rotated_vector = vector_rotate(x0=dat[:, 0], y0=dat[:, 1], z0=dat[:, 2],
                                       nx=rot_axis[:, 0], ny=rot_axis[:, 1], nz=rot_axis[:, 2],
                                       theta=-sgix2ssix_angle + spphase)
    else:  # DSI --> SGI (spin)
        print('DSI --> SGI')
        coord_out = 'sgi'
        rotated_vector = vector_rotate(x0=dat[:, 0], y0=dat[:, 1], z0=dat[:, 2],
                                       nx=rot_axis[:, 0], ny=rot_axis[:,1], nz=rot_axis[:, 2],
                                       theta=-1.*(-sgix2ssix_angle + spphase))

    store_data(name_out, data={'x': time_array,
               'y': rotated_vector}, attr_dict=dl_in)
    options(name_out, 'ytitle', '\n'.join(name_out.split('_')))



from pyspedas.projects.mms.mms_load_data import mms_load_data
from pyspedas.projects.mms.scm_tools.mms_scm_set_metadata import mms_scm_set_metadata
from pyspedas.projects.mms.mms_config import CONFIG



def mms_load_scm(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy', 
    level='l2', datatype='', varformat=None, varnames=[], suffix='', get_support_data=False,
    time_clip=True, no_update=False, available=False, notplot=False, latest_version=False, 
    major_version=False, min_version=None, cdf_version=None, spdf=False, always_prompt=False):
    """
    Load data from the MMS Search Coil Magnetometer (SCM)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2015-10-16', '2015-10-17']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4'].
            Default: '1'

        data_rate : str or list of str
            instrument data rates for SCM include ['brst' 'fast' 'slow' 'srvy'].
            Default: 'srvy'

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'
            Default: 'l2'

        datatype : str or list of str
            Valid datatypes for SCM are: ['scsrvy', 'cal', 'scb', 'scf', 'schb', 'scm', 'scs']
            Default is 'scsrvy' for srvy data, and 'scb' for brst data.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".
            Default: False

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            Default: True
            
        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
            Default: None (all variables are loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded
            Default: [] (all variables are loaded)

        suffix: str
            The tplot variable names will be given this suffix.
            Default: None

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multidimensional data products)
            Default: False

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested parameters
            Default: False

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten
            Default: False

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')
            Default: None

        min_version: str
            Specify a minimum CDF version # to load
            Default: None

        latest_version: bool
            Only grab the latest CDF version in the requested time interval
            Default: False

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval
            Default: False

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidentally save an incorrect password, or if your SDC password has changed
            Default: False

        spdf: bool
            If True, download the data from the SPDF instead of the SDC

    Returns
    ---------
        list of str
            List of tplot variables created.

    Example
    -------

        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> scm_vars = pyspedas.projects.mms.mms_load_scm(trange=['2015-10-16', '2015-10-17'])
        >>> tplot('mms1_scm_acb_gse_scsrvy_srvy_l2')

    """
    if not isinstance(data_rate, list):
        data_rate = list([data_rate])

    if isinstance(datatype, str) and datatype == '':
        # guess from data_rate
        datatype = list()
        for dr in data_rate:
            if dr == 'srvy':
                datatype.append('scsrvy')
            if dr == 'brst':
                datatype.extend(['scb', 'schb'])
        datatype = list(set(datatype)) # make it unique
    else:
        if not isinstance(datatype, list):
            datatype = list([datatype])
        # ensure datatype does not contain empty string
        datatype = list(set([dt.strip() for dt in datatype]))
        if '' in datatype:
            datatype.remove('')

    tvars = mms_load_data(trange=trange, notplot=notplot, probe=probe, data_rate=data_rate, level=level, instrument='scm',
            datatype=datatype, varformat=varformat, varnames=varnames, suffix=suffix, get_support_data=get_support_data,
            time_clip=time_clip, no_update=no_update, available=available, latest_version=latest_version, 
            major_version=major_version, min_version=min_version, cdf_version=cdf_version, spdf=spdf, 
            always_prompt=always_prompt)

    if tvars is None or available or notplot or CONFIG['download_only']:
        return tvars

    coord = ''
    if level == 'l1a':
        coord = '123'
    elif level == 'l1b':
        coord = 'scm123'
    elif level == 'l2':
        coord = 'gse'

    if not isinstance(probe, list):
        probe = [probe]

    if not isinstance(datatype, list):
        datatype = [datatype]

    probe = [str(p) for p in probe]

    for p in probe:
        for dtype in datatype:
            mms_scm_set_metadata(tvars, p, dtype, coord, suffix=suffix)

    return tvars



from pyspedas import find_datasets


def datasets(instrument=None, label=True):
    """
    Retrieves available datasets for the specified instrument on the Van Allen Probes (RBSP) mission.

    Parameters
    ----------
    instrument : str, optional
        Name of the instrument for which to find datasets. If None, finds datasets for all instruments on the mission.

    label : bool, default=True
        If True, the function prints both the dataset ID and label. If False, only the dataset ID is printed.

    Returns
    -------
    list of str
        List of available datasets for the specified instrument or for all instruments if no instrument is specified.

    Examples
    --------
    >>> pyspedas.projects.rbsp.find_datasets(instrument='REPT', label=True)
    ...
    RBSPA_REL03_ECT-REPT-SCI-L3: RBSP/ECT REPT Pitch Angle Resolved Electron and Proton Fluxes. Electron energies: 2 - 59.45 MeV. Proton energies: 21.25 - 0 MeV - D. Baker (University of Colorado at Boulder)
    ...
    """
    return find_datasets(mission='Van Allen Probes (RBSP)', instrument=instrument, label=label)




from pyspedas.tplot_tools import subtract_average
import logging

def subtract_median(
        names,
        newname=None,
        suffix=None,
        overwrite=None
):
    """
    Subtracts the median from data.

    Parameters
    ----------
    names: str/list of str
        List of pyspedas names.
    newname: str/list of str, optional
        List of new names for tplot variables.
        Default: None. If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply.
        Default: '-d'.
    overwrite: bool, optional
        If set, then tplot variables are replaced.
        Default: None

    Returns
    -------
    list of str
        Returns a list of new tplot variables created

    Examples
    --------
        >>> from pyspedas import subtract_median
        >>> pyspedas.store_data('a', data={'x':[0,4,8,12,16], 'y':[1.,2.,3.,4.,5.]})
        >>> pyspedas.subtract_median('a')

    """

    return subtract_average(names, newname=newname, suffix=suffix, overwrite=overwrite,
                     median=1)



"""
Tests for cotrans_tools functions.

These tests include the function in the following files:
    utilites/contrans.py
    utilities/cotrans_lib.py
    utilites/igrf.py
    utilites/j2000.py
    themis/cotrans_tools/dsl2gse.py

"""

import os
import unittest
import logging
import numpy as np
from numpy.testing import assert_allclose
import pyspedas
from pyspedas import (
    sm2mlt,
    tplot_copy,
    set_units,
    get_units,
    time_double,
    get_data,
    store_data,
    del_data,
    get_coords,
    set_coords,
    tplot_restore,
    tplot_names,
    options,
    subtract,
    tplot,
    cotrans,
    fac_matrix_make,
)
from pyspedas.projects.themis.cotrans.dsl2gse import dsl2gse
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
cotrans_dir = "cotrans_tools"
save_dir = os.path.join(output_dir, cotrans_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


class CotransTestCases(unittest.TestCase):
    """Tests for cotrans."""

    def test_fac_matrix_make(self):
        with self.assertLogs(level="ERROR") as log:
            doesntexist = fac_matrix_make("doesnt_exist")
            self.assertTrue(doesntexist is None)
            self.assertIn("Error reading tplot variable: doesnt_exist", log.output[0])

    def test_get_set_coords(self):
        """Test for pytplot.set_coords/get_coords"""

        del_data()
        doesntexist = get_coords("test_coord")
        self.assertTrue(doesntexist is None)
        store_data("test_coord", data={"x": [1, 2, 3, 4, 5], "y": [1, 1, 1, 1, 1]})
        with self.assertLogs(level="ERROR") as log:
            cotrans(name_in="test_coord", coord_out="geo")
            self.assertIn(
                "cotrans error: No input coordinates were provided.", log.output[0]
            )
        before = get_coords("test_coord")
        self.assertTrue(before is None)
        setcoord = set_coords("test_coord", "GSE")
        self.assertTrue(setcoord)
        after = get_coords("test_coord")
        self.assertTrue(after == "GSE")
        md = get_data("test_coord", metadata=True)
        md["data_att"]["units"] = "km"
        setcoord = set_coords("test_coord", "GSM")
        self.assertTrue(setcoord)
        md_after = get_data("test_coord", metadata=True)
        after = get_coords("test_coord")
        self.assertTrue(after == "GSM")
        self.assertTrue(md_after["data_att"]["units"] == "km")
        setcoord = set_coords("doesnt_exist", "GSM")

    def test_get_set_units(self):
        """Test for pytplot.set_coords/get_coords"""

        del_data()
        doesntexist = get_units("test_units")
        self.assertTrue(doesntexist is None)
        store_data("test_units", data={"x": [1, 2, 3, 4, 5], "y": [1, 1, 1, 1, 1]})
        before = get_units("test_units")
        self.assertTrue(before is None)
        setunits = set_units("test_units", "Km")
        self.assertTrue(setunits)
        after = get_units("test_units")
        self.assertTrue(after == "Km")
        set_coords("test_units", "GEO")
        setunits = set_units("test_units", "mm")
        self.assertTrue(setunits)
        coords_after = get_coords("test_units")
        units_after = get_units("test_units")
        self.assertTrue(coords_after == "GEO")
        self.assertTrue(units_after == "mm")

    def test_dsl2gse(self):
        """Test themis.cotrans_tools.dsl2gse."""
        del_data()
        # Try with missing variables. It should exit without problems.
        with self.assertLogs(level="ERROR") as log:
            dsl2gse("tha_fgl_dsl", "tha_fgl_gse")
            self.assertIn("Variables needed: ['tha_fgl_dsl']", log.output[0])
            self.assertIn("Variables missing: ['tha_fgl_dsl']", log.output[1])
            self.assertIn("Please load missing variables.", log.output[2])
        # Now load the needed variables.
        time_range = ["2017-03-23 00:00:00", "2017-03-23 23:59:59"]
        pyspedas.projects.themis.state(
            probe="a",
            trange=time_range,
            get_support_data=True,
            varnames=["tha_spinras", "tha_spindec"],
        )
        pyspedas.projects.themis.fgm(
            probe="a", trange=time_range, varnames=["tha_fgl_dsl"]
        )

        fac_matrix_make("tha_fgl_dsl")

        dsl2gse("tha_fgl_dsl", "tha_fgl_gse")

        t, d = get_data("tha_fgl_gse")
        # Now test the inverse.
        dsl2gse("tha_fgl_dsl", "tha_fgl_gse", isgsetodsl=True)

        self.assertTrue(abs(d[0].tolist()[0] - 15.905078404701147) <= 1e-6)
        self.assertTrue(abs(d[0].tolist()[1] - -13.962618931740064) <= 1e-6)
        self.assertTrue(abs(d[0].tolist()[2] - 16.392516225582813) <= 1e-6)

        self.assertTrue(abs(d[50000].tolist()[0] - 16.079111468932435) <= 1e-6)
        self.assertTrue(abs(d[50000].tolist()[1] - -18.858874541698583) <= 1e-6)
        self.assertTrue(abs(d[50000].tolist()[2] - 14.75796300561617) <= 1e-6)

    def test_cotrans(self):
        """Test cotrans.py GEI->GEO with Themis data."""
        del_data()
        trange = ["2010-02-25/00:00:00", "2010-02-25/23:59:59"]
        probe = "a"
        name_in = "tha_pos"
        name_out = "tha_pos_new_geo"
        pyspedas.projects.themis.state(
            probe=probe, trange=trange, time_clip=True, varnames=[name_in]
        )
        cotrans(name_in=name_in, name_out=name_out, coord_in="gei", coord_out="geo")
        din = get_data(name_in)
        in_len = len(din[0])
        dout = get_data(name_out)
        out_len = len(dout[0])
        cotrans(name_in=name_in, coord_in="gei", coord_out="geo")
        self.assertTrue(out_len == in_len)

    def test_cotrans_coord_mismatch(self):
        """Test that cotrans rejects a request where in_coord does not match the system from the variable metadata."""
        del_data()
        trange = ["2010-02-25/00:00:00", "2010-02-25/23:59:59"]
        probe = "a"
        name_in = "tha_pos"
        name_out = "tha_pos_new_geo"
        pyspedas.projects.themis.state(
            probe=probe, trange=trange, time_clip=True, varnames=[name_in]
        )
        # Metadata coordinate system is GEI, but requesting GSM->GEO transform.  This should generate an error message
        # and return failure.
        with self.assertLogs(level="ERROR"):
            result = cotrans(
                name_in=name_in, name_out=name_out, coord_in="gsm", coord_out="geo"
            )
            self.assertTrue(result == 0)

    def test_cotrans_igrf(self):
        """Test GSE->GSM and IGRF."""
        del_data()
        d = [
            [245.0, -102.0, 251.0],
            [775.0, 10.0, -10],
            [121.0, 545.0, -1.0],
            [304.65, -205.3, 856.1],
            [464.34, -561.55, -356.22],
        ]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        name_out = "tname_out"
        cotrans(
            name_out=name_out, time_in=t, data_in=d, coord_in="gse", coord_out="gsm"
        )
        in_len = len(t)
        dout = get_data(name_out)
        out_len = len(dout[0])
        gsm = dout[1][1]
        res = [775.0, 11.70357713, -7.93890939]
        self.assertTrue(out_len == in_len)
        self.assertTrue(abs(gsm[0] - res[0]) <= 1e-6)
        self.assertTrue(abs(gsm[1] - res[1]) <= 1e-6)
        self.assertTrue(abs(gsm[2] - res[2]) <= 1e-6)

    def test_cotrans_j2000(self):
        """Test GEI->J2000 and J2000 params."""
        del_data()
        d = [
            [245.0, -102.0, 251.0],
            [775.0, 10.0, -10],
            [121.0, 545.0, -1.0],
            [304.65, -205.3, 856.1],
            [464.34, -561.55, -356.22],
        ]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        name_out = "tname_out"
        cotrans(
            name_out=name_out, time_in=t, data_in=d, coord_in="gei", coord_out="j2000"
        )
        in_len = len(t)
        dout = get_data(name_out)
        out_len = len(dout[0])
        j2000 = dout[1][1]
        res = [775.01595209, 6.59521058, -11.47942543]
        self.assertTrue(out_len == in_len)
        self.assertTrue(abs(j2000[0] - res[0]) <= 1e-6)
        self.assertTrue(abs(j2000[1] - res[1]) <= 1e-6)
        self.assertTrue(abs(j2000[2] - res[2]) <= 1e-6)

    def test_cotrans_geigsm(self):
        """Test daisy chain of multiple transofrmations, GEI->GSE->GSM."""
        del_data()
        d = [
            [245.0, -102.0, 251.0],
            [775.0, 10.0, -10],
            [121.0, 545.0, -1.0],
            [304.65, -205.3, 856.1],
            [464.34, -561.55, -356.22],
        ]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        name_out = "tname_out"
        cotrans(
            name_out=name_out, time_in=t, data_in=d, coord_in="gei", coord_out="gsm"
        )
        in_len = len(t)
        dout = get_data(name_out)
        out_len = len(dout[0])
        gsm = dout[1][1]
        res = [45.98475132, 762.29816225, 132.7098884]
        self.assertTrue(out_len == in_len)
        self.assertTrue(abs(gsm[0] - res[0]) <= 1e-6)
        self.assertTrue(abs(gsm[1] - res[1]) <= 1e-6)
        self.assertTrue(abs(gsm[2] - res[2]) <= 1e-6)

    def test_all_cotrans(self):
        """Test all cotrans pairs.

        Apply transformation, then inverse transformation and compare.
        """
        with self.assertLogs(level="ERROR") as log:
            cotrans()
            self.assertIn(
                "cotrans error: No output coordinates were provided.", log.output[0]
            )
        all_cotrans = ["gei", "geo", "j2000", "gsm", "mag", "gse", "sm"]
        d = [
            [245.0, -102.0, 251.0],
            [775.0, 10.0, -10],
            [121.0, 545.0, -1.0],
            [304.65, -205.3, 856.1],
            [464.34, -561.55, -356.22],
        ]
        dd1 = d[1]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        in_len = len(t)
        name1 = "name1"
        name2 = "name2"
        count = 0
        # Test non-existent systems.
        with self.assertLogs(level="ERROR") as log:
            result = cotrans(
                name_out=name1,
                time_in=t,
                data_in=d,
                coord_in="badcoord",
                coord_out="gei",
            )
            self.assertTrue(result == 0)
            self.assertIn(
                "cotrans error: Requested input coordinate system", log.output[0]
            )
            result = cotrans(
                name_out=name1,
                time_in=t,
                data_in=d,
                coord_in="gei",
                coord_out="badcoord",
            )
            self.assertTrue(result == 0)
            self.assertIn(
                "cotrans error: Requested output coordinate system", log.output[1]
            )

        # Test empty data.
        with self.assertLogs(level="ERROR") as log:
            cotrans(
                name_out=name1, time_in=t, data_in=[], coord_in="gei", coord_out="geo"
            )
            cotrans(time_in=t, data_in=d, coord_in="gse", coord_out="gsm")
            self.assertIn("cotrans error: Data is empty.", log.output[0])
            self.assertEqual(len(log.output), 1)
        # Test all combinations.
        del_data()  # Ensure there are no leftover variables.
        with self.assertLogs(level="WARNING") as log:
            for i, coord_in in enumerate(all_cotrans):
                for j, coord_out in enumerate(all_cotrans):
                    count += 1
                    cotrans(
                        name_out=name1,
                        time_in=t,
                        data_in=d,
                        coord_in=coord_in,
                        coord_out=coord_out,
                    )
                    dout = get_data(name1)
                    out_len1 = len(dout[0])
                    self.assertTrue(out_len1 == in_len)
                    # Now perform inverse transformation.
                    cotrans(
                        name_in=name1,
                        name_out=name2,
                        coord_in=coord_out,
                        coord_out=coord_in,
                    )
                    dout2 = get_data(name2)
                    out_len2 = len(dout2[0])
                    dd2 = dout2[1][1]
                    logging.info("%d --- in: %s out: %s", count, coord_in, coord_out)
                    # print(dout[1][1])
                    # print(dd2)
                    self.assertTrue(out_len2 == in_len)
                    self.assertTrue(abs(dd1[0] - dd2[0]) <= 1e-6)
                    self.assertTrue(abs(dd1[1] - dd2[1]) <= 1e-6)
                    self.assertTrue(abs(dd1[2] - dd2[2]) <= 1e-6)
                    if coord_out == coord_in:
                        self.assertIn(
                            "Warning: coord_in equal to coord_out.",
                            log.output[i * 2],
                        )
                        self.assertIn(
                            "Warning: coord_in equal to coord_out.",
                            log.output[j * 2 + 1],
                        )
                    del_data()  # Clean up for next test.

    def test_mlt(self):
        """Test sm2mlt.

        Data is from goes18, 2023-01-01. [0, 100, 600, 1000]
        """

        # SM coordinates and MLT values from IDL
        sm_idl = [
            [33199.660, 25815.376, 2995.6487],
            [18814.747, 37614.243, 2990.7136],
            [-40428.563, -11607.452, 2995.3227],
            [15798.372, -38974.509, 2995.0028],
        ]
        mlt_idl = [14.524527, 16.228377, 1.0679544, 7.4710163]

        x = [item[0] for item in sm_idl]
        y = [item[1] for item in sm_idl]
        z = [item[2] for item in sm_idl]

        mlt_python = sm2mlt(x, y, z)

        for i in range(4):
            self.assertTrue(abs(mlt_idl[i] - mlt_python[i]) <= 1e-6)

    def test_units_coords(self):
        """Test setting units and coordinate systems (including wildcards)"""

        test_dat = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        test_times = np.array([1.0, 2.0, 3.0, 4.0, 5.0])

        store_data("v1", data={"x": test_times, "y": test_dat})
        tplot_copy("v1", "v2")
        tplot_copy("v1", "v3")
        tplot_copy("v1", "v4")

        set_coords("v?", "GSE")
        c1 = get_coords("v1")
        c2 = get_coords("v2")
        c3 = get_coords("v3")
        c4 = get_coords("v4")
        self.assertTrue(c1 == "GSE")
        self.assertTrue(c2 == "GSE")
        self.assertTrue(c3 == "GSE")
        self.assertTrue(c4 == "GSE")

        set_units("v?", "nT")
        u1 = get_units("v1")
        u2 = get_units("v2")
        u3 = get_units("v3")
        u4 = get_units("v4")
        self.assertTrue(u1 == "nT")
        self.assertTrue(u2 == "nT")
        self.assertTrue(u3 == "nT")
        self.assertTrue(u4 == "nT")

        res = set_coords("nonexistent", "GSE")
        self.assertFalse(res)

        res = set_units("xxx yyy zzz", "nT")
        self.assertFalse(res)

    def test_geogsm_geopack(self):
        # Originally reported by Bob Weigel -- PySPEDAS GEO->GSM cotrans had weird artifacts at year boundaries
        # compared to Geopack GEO->GSM transform and other Python packages
        # The SPEDAS script that creates the file: general/tools/python_validate/cotrans_gsm_test.pro

        # Load validation variables from the test file
        filename = test_data_download_file(
            validation_dir, cotrans_dir, "geogsm_gp_test.tplot", save_dir
        )

        tplot_restore(filename)
        tplot_names()
        cotrans("zgeo", "zgeo_gsm_py", coord_in="geo", coord_out="gsm")
        options("zgeo_gsm_py", "ytitle", "zgeo_gsm_py")
        subtract("zgeo_gsm_py", "zgeo_gsm", newname="pydiff")
        options("pydiff", "ytitle", "idl_python_diff")
        subtract("zgeo_gsm_gp", "zgeo_gsm", newname="gpdiff")
        options("gpdiff", "ytitle", "idl_geopack_diff")
        local_png = os.path.join(save_dir, "geo_gsm_diffs.png")
        tplot(
            ["zgeo", "zgeo_gsm", "zgeo_gsm_py", "pydiff", "gpdiff"],
            display=global_display,
            save_png=local_png,
        )
        didl = get_data("zgeo_gsm")
        dpy = get_data("zgeo_gsm_py")
        # Max relative difference around 2.7e-07
        assert_allclose(didl.y, dpy.y, atol=3.0e-07)

    def test_contrans_single_point(self):
        """Test cotrans with a single data point input."""
        data_in = np.array([[3.46410162, 3.46410162, 3.46410162]])
        time_in = [time_double("2015-12-30T00:00:00")]
        known_frames = ["GSE", "GSM", "GEI", "SM", "GEO", "J2000"]
        for f1 in known_frames:
            for f2 in known_frames:
                res = cotrans(
                    time_in=time_in, data_in=data_in, coord_in=f1, coord_out=f2
                )
                self.assertEqual(len(res[0]), 3)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from pyspedas import tnames
from pyspedas import clip, get_data, options, ylim

from ..load import load
from .remove_duplicated_tframe import remove_duplicated_tframe
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def orb(
    trange: List[str] = ['2017-03-27', '2017-03-28'],
    datatype: str = 'def',
    level: str = 'l2',
    model: str = "op",
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    version: Optional[str] = None,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads orbit data from the Arase mission

    Parameters
    ----------

        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-03-27', '2017-03-28']

        datatype: str
            Data type; Valid 'l2' options: "pre", "spre", "mpre", "lpre", "def"
            Default: 'def'

        level: str
            Data level; Valid options: 'l2', 'l3'
            Default: 'l2'

        model: str
            Field model to use for 'l3' data.  Valid options: 'op', 't89', 'ts04'
            Default: 'op'

        suffix: str
            The tplot variable names will be given this suffix.  Default: None

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
            Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded.  Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        version: str
            Set this value to specify the version of cdf files (such as "v03")
            Default: False

        ror: bool
            If true, print PI info and rules of the road.  Default: True

        uname: str
            User name. Default: None

        passed: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> orb_vars = pyspedas.projects.erg.orb(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_orb_l2_pos_gse')

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    file_res = 3600. * 24

    prefix = 'erg_orb_'+level+'_'

    if (datatype in ["pre", "spre", "mpre", "lpre"]) and (level == 'l2'):
        prefix = 'erg_orb_'+datatype+'_' + level+'_'

    if level == 'l3':
        if model == 'op':
            pathformat = 'satellite/erg/orb/'+level + \
                '/opq/%Y/%m/erg_orb_'+level+'_op_%Y%m%d_'
        else:
            pathformat = 'satellite/erg/orb/'+level+'/'+model + \
                '/%Y/%m/erg_orb_'+level+'_'+model+'_%Y%m%d_'
    elif level == 'l2':
        if datatype == 'def':
            pathformat = 'satellite/erg/orb/' + datatype + '/%Y/erg_orb_'+level+'_%Y%m%d_'
        else:
            pathformat = 'satellite/erg/orb/' + datatype + \
                '/%Y/erg_orb_' + datatype + '_'+level+'_%Y%m%d_'
    if version is None:
        pathformat += 'v??.cdf'
    else:
        pathformat += version + '.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, version=version, force_download=force_download)

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)
            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            if level == 'l3':
                print('Information about ERG L3 orbit')
            elif level == 'l2':
                print('Information about ERG orbit')
            print('')
            # print('PI: ', gatt['PI_NAME']) # not need?
            # print("Affiliation: "+gatt["PI_AFFILIATION"]) # not need?
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print('')
            print('Contact: erg-sc-core at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    remove_duplicated_tframe(tnames(prefix + '*pos*'))

    if (level == 'l2') and (datatype == 'def'):

        # remove -1.0e+30
        if prefix + 'pos_Lm' + suffix in loaded_data:
            clip(prefix + 'pos_Lm' + suffix, -1e+6, 1e6)
            _, bdata = get_data(prefix + 'pos_Lm' + suffix)
            ylim(prefix + 'pos_Lm' + suffix, np.nanmin(bdata), np.nanmax(bdata))

        # set labels
        options(prefix + 'pos_gse' + suffix, 'legend_names', ['X', 'Y', 'Z'])
        options(prefix + 'pos_gsm' + suffix, 'legend_names', ['X', 'Y', 'Z'])
        options(prefix + 'pos_sm' + suffix, 'legend_names', ['X', 'Y', 'Z'])

        options(prefix + 'pos_rmlatmlt' + suffix,
                'legend_names', ['Re', 'MLAT', 'MLT'])

        options(prefix + 'pos_eq' + suffix, 'legend_names', ['Req', 'MLT'])

        options(prefix + 'pos_iono_north' + suffix,
                'legend_names', ['GLAT', 'GLON'])
        options(prefix + 'pos_iono_south' + suffix,
                'legend_names', ['GLAT', 'GLON'])

        options(prefix + 'pos_blocal' + suffix,
                'legend_names', ['X', 'Y', 'Z'])

        options(prefix + 'pos_blocal_mag' + suffix,
                'legend_names', ['B(model)_at_ERG'])
        # options(prefix + 'pos_blocal_mag' + suffix, 'legend_names', ['B(model)\n_at_ERG']) # Can't break?

        options(prefix + 'pos_beq' + suffix, 'legend_names', ['X', 'Y', 'Z'])

        options(prefix + 'pos_Lm' + suffix, 'legend_names',
                ['90deg', '60deg', '30deg'])

        options(prefix + 'vel_gse' + suffix, 'legend_names',
                ['X[km/s]', 'Y[km/s]', 'Z[km/s]'])
        options(prefix + 'vel_gsm' + suffix, 'legend_names',
                ['X[km/s]', 'Y[km/s]', 'Z[km/s]'])
        options(prefix + 'vel_sm' + suffix, 'legend_names',
                ['X[km/s]', 'Y[km/s]', 'Z[km/s]'])

        # set color
        options(prefix + 'pos_gse' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'pos_gsm' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'pos_sm' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_rmlatmlt' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_blocal' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_beq' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_Lm' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'vel_gse' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'vel_gsm' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'vel_sm' + suffix, 'Color', ['b', 'g', 'r'])

        # set y axis to logscale
        options(prefix + 'pos_blocal_mag' + suffix, 'ylog', 1)

        options(prefix + 'pos_beq' + suffix, 'ylog', 1)

    elif (datatype in ["pre", "spre", "mpre", "lpre"]) and (level == 'l2'):

        # set labels
        options(prefix + 'pos_gse' + suffix, 'legend_names', ['X', 'Y', 'Z'])
        options(prefix + 'pos_gsm' + suffix, 'legend_names', ['X', 'Y', 'Z'])
        options(prefix + 'pos_sm' + suffix, 'legend_names', ['X', 'Y', 'Z'])

        options(prefix + 'pos_rmlatmlt' + suffix,
                'legend_names', ['Re', 'MLAT', 'MLT'])

        options(prefix + 'pos_eq' + suffix, 'legend_names', ['Req', 'MLT'])

        options(prefix + 'pos_iono_north' + suffix,
                'legend_names', ['GLAT', 'GLON'])
        options(prefix + 'pos_iono_south' + suffix,
                'legend_names', ['GLAT', 'GLON'])

        options(prefix + 'pos_blocal' + suffix,
                'legend_names', ['X', 'Y', 'Z'])

        options(prefix + 'pos_blocal_mag' + suffix,
                'legend_names', 'B(' + datatype + ')\n_at ERG')

        options(prefix + 'pos_beq' + suffix, 'legend_names', ['X', 'Y', 'Z'])

        options(prefix + 'pos_Lm' + suffix, 'legend_names',
                ['90deg', '60deg', '30deg'])

        options(prefix + 'vel_gse' + suffix, 'legend_names',
                ['X[km/s]', 'Y[km/s]', 'Z[km/s]'])
        options(prefix + 'vel_gsm' + suffix, 'legend_names',
                ['X[km/s]', 'Y[km/s]', 'Z[km/s]'])
        options(prefix + 'vel_sm' + suffix, 'legend_names',
                ['X[km/s]', 'Y[km/s]', 'Z[km/s]'])

        # set color
        options(prefix + 'pos_gse' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'pos_gsm' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'pos_sm' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_rmlatmlt' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_blocal' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_beq' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'pos_Lm' + suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'vel_gse' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'vel_gsm' + suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'vel_sm' + suffix, 'Color', ['b', 'g', 'r'])

        # set y axis to logscale
        options(prefix + 'pos_blocal_mag' + suffix, 'ylog', 1)
        options(prefix + 'pos_beq' + suffix, 'ylog', 1)

    elif level == 'l3':

        # remove -1.0e+30
        for i in range(len(loaded_data)):
            get_data_vars = get_data(loaded_data[i])
            if len(get_data_vars) < 3:
                if np.nanmin(get_data_vars[1]) < -1.0e+29:
                    clip(loaded_data[i], -1e+6, 1e6)
                    _, bdata = get_data(loaded_data[i])
                    ylim(loaded_data[i], np.nanmin(bdata), np.nanmax(bdata))

        if model in ["op", "t89", "ts04"]:

            if model == "ts04":
                model = model.upper()

            # set ytitle
            options(prefix + 'pos_lmc_' + model +
                    suffix, 'ytitle', f'Lmc ({model})')
            options(prefix + 'pos_lstar_' + model +
                    suffix, 'ytitle', f'Lstar ({model})')
            options(prefix + 'pos_I_' + model +
                    suffix, 'ytitle', f'I ({model})')
            options(prefix + 'pos_blocal_' + model +
                    suffix, 'ytitle', f'Blocal ({model})')
            options(prefix + 'pos_beq_' + model +
                    suffix, 'ytitle', f'Beq ({model})')
            options(prefix + 'pos_eq_' + model + suffix,
                    'ytitle', f'Eq_pos ({model})')
            options(prefix + 'pos_iono_north_' + model + suffix,
                    'ytitle', f'footprint_north ({model})')
            options(prefix + 'pos_iono_south_' + model + suffix,
                    'ytitle', f'footprint_south ({model})')

            # set ysubtitle
            options(prefix + 'pos_lmc_' + model + suffix,
                    'ysubtitle', '[dimensionless]')
            options(prefix + 'pos_lstar_' + model +
                    suffix, 'ysubtitle', '[dimensionless]')
            options(prefix + 'pos_I_' + model + suffix, 'ysubtitle', '[Re]')
            options(prefix + 'pos_blocal_' + model +
                    suffix, 'ysubtitle', '[nT]')
            options(prefix + 'pos_beq_' + model + suffix, 'ysubtitle', '[nT]')
            options(prefix + 'pos_eq_' + model +
                    suffix, 'ysubtitle', '[Re Hour]')
            options(prefix + 'pos_iono_north_' + model +
                    suffix, 'ysubtitle', '[deg. deg.]')
            options(prefix + 'pos_iono_south_' + model +
                    suffix, 'ysubtitle', '[deg. deg.]')

            # set ylabels
            options(prefix + 'pos_lmc_' + model + suffix, 'legend_names', ['90deg', '80deg', '70deg', '60deg', '50deg',
                                                                           '40deg', '30deg', '20deg', '10deg'])
            options(prefix + 'pos_lstar_' + model + suffix, 'legend_names', ['90deg', '80deg', '70deg', '60deg', '50deg',
                                                                             '40deg', '30deg', '20deg', '10deg'])
            options(prefix + 'pos_I_' + model + suffix, 'legend_names', ['90deg', '80deg', '70deg', '60deg', '50deg',
                                                                         '40deg', '30deg', '20deg', '10deg'])
            options(prefix + 'pos_blocal_' + model +
                    suffix, 'legend_names', '|B|')
            options(prefix + 'pos_beq_' + model +
                    suffix, 'legend_names', '|B|')
            options(prefix + 'pos_eq_' + model + suffix,
                    'legend_names', ['Re', 'MLT'])
            options(prefix + 'pos_iono_north_' + model +
                    suffix, 'legend_names', ['GLAT', 'GLON'])
            options(prefix + 'pos_iono_south_' + model +
                    suffix, 'legend_names', ['GLAT', 'GLON'])

            # set y axis to logscale
            options(prefix + 'pos_blocal_' + model + suffix, 'ylog', 1)
            options(prefix + 'pos_beq_' + model + suffix, 'ylog', 1)

    return loaded_data



import logging
import re
from pyspedas.tplot_tools import time_double, store_data, options, time_clip as tclip
from pyspedas import download, dailynames
from .kyoto_config import CONFIG


def parse_dst_html(html_text, year=None, month=None):
    """
    Parses the HTML content to extract relevant information.

    Parameters
    ----------
    html_text : str
        The HTML content to parse.
    year : int, optional
        The year to consider while parsing the HTML content. If not provided, all years are considered.
    month : int, optional
        The month to consider while parsing the HTML content. If not provided, all months are considered.

    Returns
    -------
    dict
        A dictionary containing the parsed information.
    """
    times = []
    data = []
    # remove all of the HTML before the table
    html_data = html_text[html_text.find("Hourly Equatorial Dst Values") :]
    # remove all of the HTML after the table
    html_data = html_data[: html_data.find("<!-- vvvvv S yyyymm_part3.html vvvvv -->")]
    html_lines = html_data.split("\n")
    data_strs = html_lines[5:]
    # loop over days
    for day_str in data_strs:
        # the first element of hourly_data is the day, the rest are the hourly Dst values
        hourly_data = re.findall(r'[-+]?\d+', day_str)
        ## if the data is not complete for a whole day (which is typically the case for real time data):
        if len(hourly_data[1:]) != 24:
            ## if the data is not completely missing for a whole day:
            if len(hourly_data[1:]) != 3:
                for idx, dst_value in enumerate(hourly_data[1:]):
                    ## The kyoto website uses a 4 digit format.
                    remainder = len(dst_value) % 4
                    ## if the remainder is not zero, it can be either the regular case '-23' or
                    ## the ill case '-159999'. index by 0:remainder gives the correct -23 or -15
                    if remainder > 0:
                        times.append(
                            time_double(
                                year + "-" + month + "-" + hourly_data[0] + "/" + str(idx) + ":30"
                            )
                        )
                        data.append(float(dst_value[0:remainder]))
                    ## if the remainder is zero, it can be either the regular case '-1599999' or
                    ## the ill case '9999...9999' with the number of nine being the multiple of 4.
                    ## we further test if the first four digits are 9999. If not, we simply index by
                    ## [0:4], which gives -159 in the regular case. Else, we ignore missing data.
                    elif dst_value[0:4] != '9999':
                        times.append(
                            time_double(
                                year + "-" + month + "-" + hourly_data[0] + "/" + str(idx) + ":30"
                            )
                        )
                        data.append(float(dst_value[0:4]))
        ## if the data is complete for a whole day.
        else:
            for idx, dst_value in enumerate(hourly_data[1:]):
                times.append(
                    time_double(
                        year + "-" + month + "-" + hourly_data[0] + "/" + str(idx) + ":30"
                    )
                )
                data.append(float(dst_value))

    return (times, data)

def dst(
    trange=None,
    datatypes=["final", "provisional", "realtime"],
    time_clip=True,
    remote_data_dir="http://wdc.kugi.kyoto-u.ac.jp/",
    prefix="",
    suffix="",
    no_download=False,
    local_data_dir="",
    download_only=False,
    force_download=False,
):
    """
    Loads Dst index data from the Kyoto servers.

    Parameters
    ----------
    trange : list of str, required
        Time range of interest with the format ['YYYY-MM-DD','YYYY-MM-DD'] or
        to specify more or less than a day ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Defaults to True.
    remote_data_dir : str, optional
        The remote directory from where to load the Dst index data.
        Defaults to "http://wdc.kugi.kyoto-u.ac.jp/".
    suffix : str, optional
        The tplot variable names will be given this suffix.
        By default, no suffix is added.
    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------
    list of str
        List of tplot variables created.

    Notes
    -----
    There are three types of Dst data available: final, provisional, and realtime.
    Usually, only one type is available for a particular month.
    is function tries to download final data, if this is not available then
    it downloads provisional data, and if this is not available then it downloads
    realtime data.

    Examples
    --------
    >>> from pyspedas.projects.kyoto import dst
    >>> dst_data = dst(trange=['2015-01-01', '2015-01-02'])
    >>> print(dst_data)
    kyoto_dst
    """

    vars = []  # list of tplot variables created

    if trange is None or len(trange) != 2:
        logging.error("Keyword trange with two datetimes is required to download data.")
        return vars
    if trange[0] >= trange[1]:
        logging.error("Invalid time range. End time must be greater than start time.")
        return vars

    if local_data_dir == "":
        local_data_dir = CONFIG["local_data_dir"]
    if local_data_dir[-1] != "/":
        local_data_dir += "/"

    if remote_data_dir == "":
        remote_data_dir = CONFIG["remote_data_dir_dst"]

    try:
        file_names = dailynames(file_format="%Y%m/index.html", trange=trange)
    except Exception as e:
        logging.error("Error occurred while getting file names: " + str(e))
        return vars

    # Keep unique files names only
    file_names = list(set(file_names))

    ack = """
            ******************************
            The DST data are provided by the World Data Center for Geomagnetism, Kyoto, and
            are not for redistribution (http://wdc.kugi.kyoto-u.ac.jp/). Furthermore, we thank
            the geomagnetic observatories (Kakioka [JMA], Honolulu and San Juan [USGS], Hermanus
            [RSA], Alibag [IIG]), NiCT, INTERMAGNET, and many others for their cooperation to
            make the Dst index available.
            ******************************
        """

    times = []
    data = []
    datatypes = ["final", "provisional", "realtime"]

    # Final files
    for datatype in datatypes:

        for filename in file_names:
            yyyymm = ""
            if len(filename) > 6:
                yyyymm = filename[:6]
            url = remote_data_dir + "dst_" + datatype + "/" + filename
            local_path = local_data_dir + "dst_" + datatype + "/" + yyyymm + "/"
            fname = download(
                url, local_path=local_path, no_download=no_download, text_only=True, force_download=force_download,
            )

            if download_only:
                continue  # skip to the next file

            if fname is None or len(fname) < 1:
                logging.error("Error occurred while downloading: " + url)
                continue  # skip to the next file
            try:
                with open(fname[0], "r") as file:
                    html_text = file.read()
                file_times, file_data = parse_dst_html(
                    html_text, year=filename[:4], month=filename[4:6]
                )
                times.extend(file_times)
                data.extend(file_data)
            except Exception as e:
                logging.error(
                    "Error occurred while parsing " + filename + ": " + str(e)
                )
                continue  # skip to the next file

        # At this point, we have all data for one datatype
        # If we have data, we can break the loop, otherwise we try other datatypes
        if len(times) != 0:
            break

    if len(times) == 0:
        logging.error("No data found.")
        return vars

    varname = prefix + "kyoto_dst" + suffix
    store_data(varname, data={"x": times, "y": data})
    options(varname, "ytitle", "Dst (" + datatype + ")")
    vars.append(varname)

    if time_clip:
        tclip(varname, trange[0], trange[1], overwrite=True)

    logging.info(ack)

    return vars



import fnmatch
import logging

from pyspedas.tplot_tools import data_quants
from pyspedas.tplot_tools import tplot_names

def wildcard_expand(master_list, patterns, case_sensitive=True, split_whitespace=True, quiet=False):
    """ Find elements in master list matching one or more wild card patterns

    Parameters
    ----------
    master_list: str or list of str
        One or more strings (for example, the set of all tplot variable names)
    patterns: str or list of str
        One or more patterns, which may contain wildcard characters, to be matched agains the master list
    case_sensitive: bool
        A boolean flag indicating whether case-sensitive matching should be performed. Default: True
    split_whitespace: bool
        A boolean flag indicating that master_list or patterns should be treated as space-delimited lists.  Set to False if embedded spaces
        are to be matched.  Default: True
    quiet: bool
        A boolean flag indicating whether to suppress "match not found" warnings. Default: False

    Returns
    -------
    list of str
        A list of elements from master_list that match at least one of the input patterns. Duplicates in the output will be removed.

    Examples
    --------

    >>> from pyspedas import wildcard_expand
    >>> master_list = ['mms1_mec_r_sm', 'mms2_mec_r_sm', 'mms3_mec_r_gsm', 'tha_pos_gsm']
    >>> wildcard_expand(master_list,'*')
    ['mms1_mec_r_sm', 'mms2_mec_r_sm', 'mms3_mec_r_gsm', 'tha_pos_gsm']
    >>> wildcard_expand(master_list, ['mms?_mec_r_sm'])
    ['mms1_mec_r_sm', 'mms2_mec_r_sm']
    >>>wildcard_expand(master_list, ['*SM'], case_sensitive=False)
    ['mms1_mec_r_sm', 'mms2_mec_r_sm', 'mms3_mec_r_gsm', 'tha_pos_gsm']
    >>> wildcard_expand(master_list,'tha_pos_gsm')
    ['tha_pos_gsm']

    """
    if master_list is None:
        logging.warning("wildcard_expand: master_list must be a non-empty string or list")
        return []
    if patterns is None:
        # One could argue that it makes sense to return the entire master list here?
        return []

    matched_items=set()
    ordered_matches=[]

    # Convert inputs to lists if necessary
    if isinstance(patterns, str):
        patterns=[patterns]
    if isinstance(master_list, str):
        master_list=[master_list]

    # Warn if either input is empty
    if len(master_list) == 0 or master_list is None or master_list==['']:
        logging.warning("wildcard_expand: empty master list")
        return []
    if len(patterns) == 0 or patterns is None or patterns==['']:
        logging.warning("wildcard_expand: empty pattern list")
        return []

    # If split_whitespace is true, expand master list and patterns
    if split_whitespace:
        expanded_master_list = []
        for p in master_list:
            p_expanded=p.split(' ')
            expanded_master_list.extend(p_expanded)
        master_list = expanded_master_list
        # Do the same for the patterns
        expanded_patterns=[]
        for pattern in patterns:
            split_pattern=pattern.split(' ')
            expanded_patterns.extend(split_pattern)
        patterns=expanded_patterns


    # Prepare temporary lists for matching
    if case_sensitive:
        temp_master_list = master_list
        temp_patterns = patterns
    else:
        temp_master_list = [item.lower() for item in master_list]
        temp_patterns = [pattern.lower() for pattern in patterns]

    # Check each pattern against each item in the master list using fnmatchcase
    for pattern in temp_patterns:
        found = False
        for temp_item, original_item in zip(temp_master_list, master_list):
            if fnmatch.fnmatchcase(temp_item, pattern) and original_item not in matched_items:
                found = True
                ordered_matches.append(original_item)
                matched_items.add(original_item)
            elif fnmatch.fnmatchcase(temp_item, pattern):
                found = True
        if not found and not quiet:
            logging.warning("wildcard_expand: No match found for %s", pattern)

    return ordered_matches

def tname_byindex(tvar_index):
    """ Return a tplot variable name given an integer index

    Parameters
    ----------
    tvar_index: int
        Index into the list of current tplot variable names

    Returns
    -------
    str or None
        Returns the tplot variable name, or None if the index is invalid.

    """
    namelist = list(data_quants.keys())
    if tvar_index >= len(namelist) or tvar_index < 0:
        logging.warning("tname_byindex: index value %d out of bounds", tvar_index)
        return None
    else:
        return namelist[tvar_index]

def tindex_byname(tvar_name):
    """ Return a tplot variable index given a tplot variable name

    Parameters
    ----------
    tvar_name: str
        Name of a tplot variable

    Returns
    -------
    int
        Index of the variable name, or None if not found

    """
    namelist = list(data_quants.keys())
    try:
        index = namelist.index(tvar_name)
        return index
    except ValueError:
        return None


def tplot_wildcard_expand(patterns, case_sensitive=True, quiet=False):
    """ Return tplot variable names matching one or more wildcard patterns

    Parameters
    ----------
    patterns: str, int, list(str!int)
        One or more patterns, which may contain wildcard characters, to be matched against the master list
        Integers are converted to strings using tname_byindex()
    case_sensitive: bool
        A boolean flag indicating whether case-sensitive matching should be performed. Default: True
    quiet: bool
        A boolean flag indicating "no match found" log messages should be suppressed. Default: False

    Returns
    -------
    list of str
        A list of tplot variable names that match at least one of the input patterns. Duplicate values in the output will be removed.

    """
    input_pattern_list = patterns
    if patterns is None:
        return []
    elif isinstance(patterns,str) or isinstance(patterns,int):
        input_pattern_list = [patterns]
    elif isinstance(patterns,list):
        input_pattern_list = patterns
    else:
        logging.warning("tplot_wildcard_expand: patterns must be a string, int or list")
        return []

    string_patterns = []
    # Get list of existing tplot names
    tn = tplot_names(quiet=True)

    if len(tn) == 0:
        if not quiet:
            logging.warning("tplot_wildcard_expand: List of tplot variable names is empty.")
        return []

    # Interpret integers as tplot variable indices
    for item in input_pattern_list:
        if isinstance(item, int):
            name = tname_byindex(item)
            if name is not None:
                string_patterns.append(name)
            else:
                string_patterns.append(item)
        elif isinstance(item,str):
            # tplot names can have embedded spaces, but we also want to allow space-delimited
            # lists. If a space-delimited list is provided, check to see if the un-split version is
            # a valid tplot name.  If not, try splitting it and look for the subpatterns.
            # This case used to also be triggered if there were wildcard characters present.  But this prevented
            # using space-delimited patterns with wildcards.
            if item in tn:
                string_patterns.append(item)
            elif ' ' in item:
                expanded = item.split(' ')
                string_patterns.extend(expanded)
            else:
                # No embedded spaces, and not found in list of names.
                # But there might be wildcards to expand.
                string_patterns.append(item)
        else:
            logging.warning("tplot_wildcard_expand: bad input: "+str(item) +" Patterns must be a string or int")

    return wildcard_expand(tn, patterns=string_patterns, case_sensitive=case_sensitive, split_whitespace=False, quiet=quiet)


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas
import logging


def get_ylimits(name, trg=None):
    """
    Extracts the y-limits from the Tplot Variables stored in memory.

    Parameters
    ----------
    name : str
        Name of the tplot variable.
    trg : list, optional
        The time range to look in.

    Returns
    -------
    ymin : int, float, numeric type of pyspedas data
        The minimum value of y.
    ymax : int, float, numeric type of pyspedas data
        The maximum value of y.

    Examples
    --------
    >>> import pyspedas
    >>> x_data = [1, 2, 3, 4, 5]
    >>> y_data = [1, 2, 3, 4, 5]
    >>> pyspedas.store_data("Variable1", data={'x': x_data, 'y': y_data})
    >>> y1, y2 = pyspedas.get_ylimits("Variable1")
    """

    if isinstance(name, int):
        name = list(pyspedas.tplot_tools.data_quants.keys())[name-1]
    if not isinstance(name, list):
        name = [name]
    name_num = len(name)
    ymin = None
    ymax = None

    for i in range(name_num):

        if name[i] not in pyspedas.tplot_tools.data_quants.keys():
            logging.info(str(name[i]) + " is currently not in pyspedas.")
            return
        y = pyspedas.tplot_tools.data_quants[name[i]]

        # Slice the data around a time range
        if trg is not None:
            y = y.sel(time=slice(trg[0], trg[1]))

        ymin = y.min(skipna=True).values.item()
        ymax = y.max(skipna=False).values.item()

    return ymin, ymax



# Example of getting MTH5 datasets availability

from pyspedas.mth5.utilities import datasets

# Getting a dictionary of valid datasets for a given time range
valid_dataset = datasets(trange=["2015-06-22", "2015-06-23"])

# List of all the networks
nets = valid_dataset.keys()
print("Networks: ", nets)

# List of all the stations for a given network
net = list(nets)[0]
print("Stations: ", valid_dataset[net].keys())

# Getting data availability for a given network and station in 2015
res = datasets(trange=["2015-01-01", "2016-01-01"], network="4P", station="TNU48")
print("Data availability", res)




import numpy as np
from pyspedas.tplot_tools import get_data, store_data


def tcrossp(
        v1,
        v2,
        newname=None,
        return_data=False
):
    """
    Calculates the cross product of two tplot variables

    Parameters
    ----------
    v1: str
        First tplot variable
    v2: str
        Second tplot variable
    newname: str, optional
        Name of the output variable
        Default: v1_cross_v2
    return_data: bool
        Returns the data as an ndarray instead of creating a tplot variable
        Default: False

    Returns
    --------
    str
        Name of the tplot variable created

    Examples
    --------

        >>> # Compute the cross product of two time series
        >>> import pyspedas
        >>> x1 = [0, 4, 8]
        >>> x2 = [0, 4, 8]
        >>> time1 = [pyspedas.time_float("2020-01-01") + i for i in x1]
        >>> time2 = [pyspedas.time_float("2020-01-01") + i for i in x2]
        >>> pyspedas.store_data("a", data={"x": time1, "y": [[1, 2, 3],[2, 3, 4],[3, 4, 5]]})
        >>> pyspedas.store_data("c", data={"x": time2, "y": [[1, 4, 1],[2, 5, 2],[3, 5, 3]]})
        >>> n = pyspedas.tcrossp("a", "c", newname="a_cross_c")
        >>> print('new tplot variable:', n)
        >>> ac = pyspedas.get_data(n)
        >>> print(ac)

    """

    v1_data = None
    v2_data = None

    if not isinstance(v1, np.ndarray) and isinstance(v1, str):
        v1_data = get_data(v1)
        v1_name = v1

        if v1_data is not None:
            data1 = v1_data[1]
    else:
        v1_name = 'var1'
        data1 = v1

    if not isinstance(v2, np.ndarray) and isinstance(v2, str):
        v2_data = get_data(v2)
        v2_name = v2

        if v2_data is not None:
            data2 = v2_data[1]
    else:
        v2_name = 'var2'
        data2 = v2

    if newname is None:
        newname = v1_name + '_cross_' + v2_name

    cp = np.cross(data1, data2)

    if return_data:
        return cp
    else:
        out = cp
        if v2_data is None:
            if len(cp.shape) == 1:
                out = np.atleast_2d(cp)
            times = np.zeros(out.shape[0])
        else:
            times = v2_data[0]
        store_data(newname, data={'x': times, 'y': out})
        return newname




import numpy as np
from pyspedas import clip, get_data, options, store_data, ylim, zlim
from pyspedas import tcopy

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def lepi(
    trange: List[str] = ['2017-07-01', '2017-07-02'],
    datatype: str = 'omniflux',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    version: Optional[str] = None,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the LEP-i experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-07-01', '2017-07-02']

        datatype: str
            Data type; Valid 'l2' options: '3dflux', 'omniflux'  Valid 'l3' options: 'pa'
            Default: 'omniflux'

        level: str
            Data level; Valid options: 'l2','l3'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            If Trye, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded. Default: [] (all data loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        ror: bool
            If set, print PI info and rules of the road
            Default: False

        version: str
            Set this value to specify the version of cdf files (such as "v03_00")
            Default: None

        uname: str
            User name.  Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> lepi_vars = pyspedas.projects.erg.lepi(trange=['2017-04-09', '2017-04-10'])
    >>> tplot('erg_lepi_l2_omniflux_FODO')

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    if level == 'l3':
        datatype = 'pa'

    file_res = 3600. * 24
    prefix = 'erg_lepi_'+level+'_' + datatype + '_'

    pathformat = 'satellite/erg/lepi/'+level+'/'+datatype + \
        '/%Y/%m/erg_lepi_'+level+'_'+datatype+'_%Y%m%d_'
    if version is None:
        pathformat += 'v??_??.cdf'
    else:
        pathformat += version + '.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    flag_FPDO = False # In case it doesn't get set later

    if (len(loaded_data) > 0) and ror:
        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG LEPi')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ", gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of LEPi L2: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Lepi')
            print(
                'RoR of ERG/LEPi: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Lepi#Rules_of_the_Road')
            print('')
            print('Contact: erg_lepi_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if (datatype == 'omniflux') and (level == 'l2'):
        tplot_variables = []

        if prefix + 'FPDO' + suffix in loaded_data:
            store_data(prefix + 'FPDO' + suffix,
                       newname=prefix + 'FPDO_raw' + suffix)
            loaded_data.append(prefix + 'FPDO_raw' + suffix)
            get_data_vars = get_data(prefix + 'FPDO_raw' + suffix)
            flag_FPDO = store_data(prefix + 'FPDO' + suffix, data={'x': get_data_vars[0],
                                                                   'y': get_data_vars[1][:, :-2],
                                                                   'v': get_data_vars[2][:-2]})
            tplot_variables.append(prefix + 'FPDO' + suffix)

        if prefix + 'FHEDO' + suffix in loaded_data:
            store_data(prefix + 'FHEDO' + suffix,
                       newname=prefix + 'FHEDO_raw' + suffix)
            loaded_data.append(prefix + 'FHEDO_raw' + suffix)
            get_data_vars = get_data(prefix + 'FHEDO_raw' + suffix)
            store_data(prefix + 'FHEDO' + suffix, data={'x': get_data_vars[0],
                                                        'y': get_data_vars[1][:, :-2],
                                                        'v': get_data_vars[2][:-2]})
            tplot_variables.append(prefix + 'FHEDO' + suffix)

        if prefix + 'FODO' + suffix in loaded_data:
            store_data(prefix + 'FODO' + suffix,
                       newname=prefix + 'FODO_raw' + suffix)
            loaded_data.append(prefix + 'FODO_raw' + suffix)
            get_data_vars = get_data(prefix + 'FODO_raw' + suffix)
            store_data(prefix + 'FODO' + suffix, data={'x': get_data_vars[0],
                                                       'y': get_data_vars[1][:, :-2],
                                                       'v': get_data_vars[2][:-2]})
            tplot_variables.append(prefix + 'FODO' + suffix)

        # remove minus valuse of y array
        if flag_FPDO:
            clip(prefix + 'FPDO' + suffix, 0., 2.e+16)

        # set ytitle
        options(tplot_variables, 'ytitle', 'LEPi\nomniflux\nLv2\nEnergy')

        # set ysubtitle
        options(tplot_variables, 'ysubtitle', '[keV/q]')

        # set spectrogram plot option
        options(tplot_variables, 'Spec', 1)

        # set y axis to logscale
        options(tplot_variables, 'ylog', 1)

        for i in range(len(tplot_variables)):
            # set ylim
            ylim(tplot_variables[i], 0.01, 25.0)
            # set zlim
            zlim(tplot_variables[i], 1e+1, 1e+9)

        # set ztitle
        options(tplot_variables, 'ztitle', '[/cm^2-str-s-keV]')

        # set z axis to logscale
        options(tplot_variables, 'zlog', 1)

        # change colormap option
        options(tplot_variables, 'Colormap', 'jet')

    elif (datatype == '3dflux') and (level == 'l2') and (not notplot):
        vns_fidu = [ 'FPDU',  'FPDU_sub', 'FHEDU', 'FHEDU_sub', 'FODU', 'FODU_sub'] 
        vns_fiedu = [ 'FPEDU', 'FPEDU_sub', 'FHEEDU', 'FHEEDU_sub', 'FOEDU', 'FOEDU_sub']
        vns_cnt = ['FPDU_COUNT_RAW','FPDU_COUNT_RAW_sub', 'FHEDU_COUNT_RAW','FHEDU_COUNT_RAW_sub', 'FODU_COUNT_RAW', 'FODU_COUNT_RAW_sub']
        vns_list = vns_fidu + vns_fiedu + vns_cnt

        v2_array_for_not_sub = np.arange(8)
        v2_array_for_sub = np.arange(7) + 8
        v3_array = np.arange(16)
        
        for vns_pattarn in vns_list:
            t_plot_name = prefix + vns_pattarn + suffix
            if t_plot_name in loaded_data:
                tcopy(t_plot_name, t_plot_name + '_raw')
                get_data_vars_temporal = get_data(t_plot_name)
                meta_data_in = get_data(t_plot_name, metadata=True)
                if 'sub' in t_plot_name:
                    store_data(t_plot_name,data={'x':get_data_vars_temporal[0],
                                                'y':get_data_vars_temporal[1][:,0:30,:,:],
                                                'v1':get_data_vars_temporal[2][0:30],
                                                'v2':v2_array_for_sub,
                                                'v3':v3_array},
                            attr_dict=meta_data_in)
                else:
                    store_data(t_plot_name,data={'x':get_data_vars_temporal[0],
                                                'y':get_data_vars_temporal[1][:,0:30,:,:],
                                                'v1':get_data_vars_temporal[2][0:30],
                                                'v2':v2_array_for_not_sub,
                                                'v3':v3_array},
                            attr_dict=meta_data_in)
                ylim(t_plot_name,  0.01, 30.)
                options(t_plot_name, 'ylog', 1)
                options(t_plot_name, 'zlog', 1)
        
        """comment out in order to match the result of part_product of IDL.
        if prefix + 'FPDU' + suffix in loaded_data:
            clip(prefix + 'FPDU' + suffix, -1.0e+10, 1.0e+10)
        if prefix + 'FHEDU' + suffix in loaded_data:
            clip(prefix + 'FHEDU' + suffix, -1.0e+10, 1.0e+10)
        if prefix + 'FODU' + suffix in loaded_data:
            clip(prefix + 'FODU' + suffix, -1.0e+10, 1.0e+10)"""

    elif level == 'l3':
        tplot_variables = []

        if prefix + 'FPDU' + suffix in loaded_data:
            tplot_variables.append(prefix + 'FPDU' + suffix)
            get_data_vars = get_data(prefix + 'FPDU' + suffix)
            ylim(prefix + 'FPDU' + suffix, 0, 180)
            zlim(prefix + 'FPDU' + suffix, 1e2, 1e5)
            options(prefix + 'FPDU' + suffix, 'spec', 1)
            ytitle_keV_array = np.round(np.nan_to_num(get_data_vars[2]), 2)
            for i in range(get_data_vars[1].shape[1]):
                tplot_name = prefix + 'pabin_' + \
                    str(i).zfill(2) + '_FPDU' + suffix
                store_data(tplot_name, data={'x': get_data_vars[0],
                                             'y': get_data_vars[1][:, i, :],
                                             'v': get_data_vars[3]})
                options(tplot_name, 'spec', 1)
                ylim(tplot_name, 0, 180)
                zlim(tplot_name, 1e2, 1e5)
                options(tplot_name, 'ytitle', 'ERG LEP-i P\n' +
                        str(ytitle_keV_array[i]) + ' keV\nPitch angle')
                tplot_variables.append(tplot_name)

            loaded_data += tplot_variables[1:]

        tplot_variables_length = len(tplot_variables)

        if prefix + 'FHEDU' + suffix in loaded_data:
            tplot_variables.append(prefix + 'FHEDU' + suffix)
            get_data_vars = get_data(prefix + 'FHEDU' + suffix)
            options(prefix + 'FHEDU' + suffix, 'spec', 1)
            ylim(prefix + 'FHEDU' + suffix, 0, 180)
            zlim(prefix + 'FHEDU' + suffix, 1e2, 1e5)
            ytitle_keV_array = np.round(np.nan_to_num(get_data_vars[2]), 2)
            for i in range(get_data_vars[1].shape[1]):
                tplot_name = prefix + 'pabin_' + \
                    str(i).zfill(2) + '_FHEDU' + suffix
                store_data(tplot_name, data={'x': get_data_vars[0],
                                             'y': get_data_vars[1][:, i, :],
                                             'v': get_data_vars[3]})
                options(tplot_name, 'spec', 1)
                ylim(tplot_name, 0, 180)
                zlim(tplot_name, 1e2, 1e5)
                options(tplot_name, 'ytitle', 'ERG LEP-i P\n' +
                        str(ytitle_keV_array[i]) + ' keV\nPitch angle')
                tplot_variables.append(tplot_name)

            loaded_data += tplot_variables[tplot_variables_length + 1:]

        tplot_variables_length = len(tplot_variables)

        if prefix + 'FODU' + suffix in loaded_data:
            tplot_variables.append(prefix + 'FODU' + suffix)
            get_data_vars = get_data(prefix + 'FODU' + suffix)
            options(prefix + 'FODU' + suffix, 'spec', 1)
            ylim(prefix + 'FODU' + suffix, 0, 180)
            zlim(prefix + 'FODU' + suffix, 1e2, 1e5)
            ytitle_keV_array = np.round(np.nan_to_num(get_data_vars[2]), 2)
            for i in range(get_data_vars[1].shape[1]):
                tplot_name = prefix + 'pabin_' + \
                    str(i).zfill(2) + '_FODU' + suffix
                store_data(tplot_name, data={'x': get_data_vars[0],
                                             'y': get_data_vars[1][:, i, :],
                                             'v': get_data_vars[3]})
                options(tplot_name, 'spec', 1)
                ylim(tplot_name, 0, 180)
                zlim(tplot_name, 1e2, 1e5)
                options(tplot_name, 'ytitle', 'ERG LEP-i P\n' +
                        str(ytitle_keV_array[i]) + ' keV\nPitch angle')
                tplot_variables.append(tplot_name)

            loaded_data += tplot_variables[tplot_variables_length + 1:]

        options(tplot_variables, 'zlog', 1)
        options(tplot_variables, 'ysubtitle', 'PA [deg]')
        options(tplot_variables, 'colormap', 'jet')
        options(tplot_variables, 'ztitle', '[/s-cm^{2}-sr-keV/q]')

    return loaded_data



import warnings
import logging
import numpy as np
from pyspedas import tnames
from pyspedas.tplot_tools import get_data, store_data, options
from pyspedas.projects.mms.eis_tools.mms_eis_pad_spinavg import mms_eis_pad_spinavg

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_eis_pad(scopes=['0', '1', '2', '3', '4', '5'], probe='1', level='l2', 
    data_rate='srvy', datatype='extof', species='proton', data_units='flux', 
    energy=[55, 800], size_pabin=15, suffix=''):
    """
    Calculate pitch angle distributions using data from the MMS Energetic Ion Spectrometer (EIS)
    
    Parameters
    ----------
        scopes: list of str
            telescope #s to include in the calculation

        probe: str
            probe #, e.g., '4' for MMS4

        level: str
            data level for calculation (default: 'l2')

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst' (default: 'srvy')

        datatype: str
            'extof' or 'phxtof' (default: 'extof')

        species: str
            species for calculation (default: 'proton')

        data_units: str
            'flux' or 'cps' (default: 'flux')

        energy: list of float
            energy range to include in the calculation (default: [55, 800])

        size_pabin: int
            size of the pitch angle bins, in degrees (default: 15)

        suffix: str
            suffix of the loaded data

    Returns
    ----------
        Name of tplot variables created.
    """

    # allow for the user to input probe, datatype and species as a string or a list of strings
    if not isinstance(probe, list): probe = [probe]
    if not isinstance(datatype, list): datatype = [datatype]
    if not isinstance(species, list): species = [species]

    if data_units == 'cps':
        units_label = '1/s'
    else:
        units_label = '1/(cm^2-sr-s-keV)'

    if len(scopes) == 1:
        scope_suffix = '_t' + scopes + suffix
    elif len(scopes) == 6:
        scope_suffix = '_omni' + suffix

    # set up the number of pa bins to create
    n_pabins = 180./size_pabin
    pa_label = [180.*n_pabin/n_pabins+size_pabin/2. for n_pabin in range(0, int(n_pabins))]

    # Account for angular response (finite field of view) of instruments
    pa_halfang_width = 10.0 # deg
    delta_pa = size_pabin/2.

    out_vars = []

    # the probes will need to be strings beyond this point
    probe = [str(p) for p in probe]

    logging.info('Calculating the EIS pitch angle distribution; this may take several minutes')

    for probe_id in probe:
        prefix = 'mms' + probe_id + '_epd_eis_' + data_rate + '_' + level + '_'

        for datatype_id in datatype:
            pa_data = get_data(prefix + datatype_id + '_pitch_angle_t0' + suffix)
            if pa_data is None:
                logging.error('No ' + data_rate + ' ' + datatype_id + ' data is currently loaded for MMS' + probe_id + ' for the selected time period')
                return

            for species_id in species:
                pa_times, pa_data = get_data(prefix + datatype_id + '_pitch_angle_t0' + suffix)

                pa_file = np.zeros([len(pa_times), len(scopes)])

                omni_times, omni_data, omni_energies = get_data(prefix + datatype_id + '_' + species_id + '_' + data_units + '_omni' + suffix)
                erange = get_data(prefix + datatype_id + '_' + species_id + '_energy_range' + suffix)
                inchan_check_low = np.zeros(len(omni_energies))
                inchan_check_hi = np.zeros(len(omni_energies))
                inchan_check_low[np.where((erange[:, 0] >= energy[0]) & (erange[:, 0] <= energy[1]))[0]] = 1
                inchan_check_hi[np.where((erange[:, 1] >= energy[0]) & (erange[:, 1] <= energy[1]))[0]] = 1
                these_energies = np.where((inchan_check_low > 0) | (inchan_check_hi > 0))[0]

                if sum(inchan_check_low) == 0 or sum(inchan_check_hi) == 0:
                    logging.error( 'Energy range selected is not covered by the detector for ' + datatype_id + ' ' + species_id + ' ' + data_units)
                    continue

                flux_file = np.zeros([len(pa_times), len(scopes), len(these_energies)])
                flux_file[:] = 'nan'
                pa_flux = np.zeros([len(pa_times), int(n_pabins), len(these_energies)])
                pa_flux[:] = 'nan'
                pa_num_in_bin = np.zeros([len(pa_times), int(n_pabins), len(these_energies)])

                for t, scope in enumerate(scopes):
                    pa_times, pa_data = get_data(prefix + datatype_id + '_pitch_angle_t' + scope + suffix)

                    pa_file[:, t] = pa_data

                    # use wild cards to figure out what this variable name should be for telescope 0
                    this_variable = tnames(prefix + datatype_id + '_' + species_id + '*_' + data_units + '_t0' + suffix)

                    if level == 'l2' or level == 'l1b':
                        pvalue = this_variable[0].split('_')[7]
                    else:
                        pvalue = ''

                    # get flux from each detector
                    flux_times, flux_data, flux_energies = get_data(prefix + datatype_id + '_' + species_id + '_' + pvalue + '_' + data_units + '_t' + scope + suffix)

                    # get energy range of interest
                    e = flux_energies[these_energies]

                    flux_file[:, t, :] = flux_data[:, these_energies]

                # CREATE PAD VARIABLES FOR EACH ENERGY CHANNEL IN USER-DEFINED ENERGY RANGE
                for i, flux_time in enumerate(flux_times):
                    for j, pa_bin in enumerate(range(0, int(n_pabins))):
                        for ee in range(0, len(these_energies)):
                            ind = np.where((pa_file[i, :] + pa_halfang_width >= pa_label[j] - delta_pa) & (pa_file[i, :] - pa_halfang_width < pa_label[j] + delta_pa))[0]
                            if ind.size != 0:
                                pa_flux[i, j, ee] = nanmean(flux_file[i, ind, ee], axis=0)

                for ee in range(0, len(these_energies)):
                    # energy_string = str(int(flux_energies[these_energies[ee]])) + 'keV'
                    energy_string = str(int(erange[these_energies[ee], 0])) + '_' + str(int(erange[these_energies[ee], 1])) + 'keV'
                    new_name = prefix + datatype_id + '_' + energy_string + '_' + species_id + '_' + data_units + scope_suffix + '_pad'
                    store_data(new_name, data={'x': flux_times, 'y': pa_flux[:, :, ee], 'v': pa_label})
                    options(new_name, 'ylog', False)
                    options(new_name, 'zlog', True)
                    options(new_name, 'spec', True)
                    options(new_name, 'ztitle', units_label)
                    options(new_name, 'ytitle', 'MMS' + str(probe_id) + ' ' + datatype_id + ' PA (deg)')
                    out_vars.append(new_name)

                try:
                    store_data(prefix + datatype_id + '_' + species_id + '_' + data_units + scope_suffix + '_pads', data={'x': flux_times, 'y': pa_flux, 'v1': pa_label, 'v2': omni_energies[these_energies]})
                    out_vars.append(prefix + datatype_id + '_' + species_id + '_' + data_units + scope_suffix + '_pads')
                except ValueError: # kludge to avoid crash in case of single energy
                    logging.error('Problem creating: ' + prefix + datatype_id + '_' + species_id + '_' + data_units + scope_suffix + '_pads')

                # only create the integral PAD variable if the user defined energy range covers more than 1 EIS energy channel
                if these_energies.size == 1:
                    continue

                # CREATE PAD VARIABLE INTEGRATED OVER USER-DEFINED ENERGY RANGE
                # energy_range_string = str(int(flux_energies[these_energies[0]])) + '-' + str(int(flux_energies[these_energies[-1]])) + 'keV'
                energy_range_string = str(int(erange[these_energies[0], 0])) + '-' + str(int(erange[these_energies[-1], 1])) + 'keV'
                new_name = prefix + datatype_id + '_' + energy_range_string + '_' + species_id + '_' + data_units + scope_suffix + '_pad'

                avg_pa_flux = np.zeros([len(flux_times), int(n_pabins)])
                avg_pa_flux[:] = 'nan'

                for tt in range(0, len(flux_times)):
                    for bb in range(0, int(n_pabins)):
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore", category=RuntimeWarning)
                            avg_pa_flux[tt, bb] = nanmean(pa_flux[tt, bb, :])

                store_data(new_name, data={'x': flux_times, 'y': avg_pa_flux, 'v': pa_label})
                options(new_name, 'ylog', False)
                options(new_name, 'zlog', True)
                options(new_name, 'spec', True)
                options(new_name, 'ztitle', units_label)
                options(new_name, 'ytitle', 'MMS' + str(probe_id) + ' ' + datatype_id + ' PA')
                options(new_name, 'ysubtitle', '[deg]')
                out_vars.append(new_name)

                spin_avg_pads = mms_eis_pad_spinavg(scopes=scopes, probe=probe_id, data_rate=data_rate, level=level, datatype=datatype_id, data_units=data_units, species=species_id, energy=energy, size_pabin=size_pabin, suffix=suffix)

                for spin_avg_pad in spin_avg_pads:
                    out_vars.append(spin_avg_pad)

    return out_vars



import logging
import numpy as np
import pyspedas
#from pyspedas import tinterpol, interpol

from pyspedas.tplot_tools import get_coords
from pyspedas.tplot_tools import tnormalize
from pyspedas.tplot_tools import tcrossp

from pyspedas.tplot_tools import get_data, store_data, del_data

from pyspedas.cotrans_tools.xyz_to_polar import xyz_to_polar
from pyspedas.cotrans_tools.cotrans import cotrans
from .rotmat_set_coords import rotmat_set_coords

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


# === Helper Functions ===

def validate_vector_data(var_name, expected_dim=3):
    """
    Retrieve tplot variable and check that the 'y' field is an Nx{expected_dim} array.
    """
    data = get_data(var_name)
    if data is None:
        logging.error(f"Error reading tplot variable: {var_name}")
        return None
    if data.y.ndim != 2 or data.y.shape[1] != expected_dim:
        logging.error(f"{var_name} data must be an Nx{expected_dim} array")
        return None
    return data


def interpolate_position(pos_var_name, interp_to):
    """
    Interpolate position data to the time grid of the magnetic field variable.
    The result is stored as pos_var_name + "-itrp".
    """
    pyspedas.tinterpol(names=pos_var_name, interp_to=interp_to, method="linear",
              newname=None, extrapolate=True, suffix="-itrp")
    interp_var = pos_var_name + "-itrp"
    interp_data = get_data(interp_var)
    if interp_data is None:
        logging.error("Interpolation failed for %s", pos_var_name)
    return interp_data


def normalize_vectors(vectors):
    """Normalize an array of vectors along the last axis."""
    norms = np.linalg.norm(vectors, axis=1, keepdims=True)
    norms[norms == 0] = 1  # avoid division by zero
    return vectors / norms


def pos_cotrans(pos_times, pos_data, mag_var_name):
    """Coordinate transformation of position"""
    store_data('fac_mat_pos_tmp', data={'x': pos_times, 'y': pos_data})

    # Retrieve the coordinate system mad_var_name
    coord_system = get_coords(mag_var_name)

    cotrans(name_in='fac_mat_pos_tmp', coord_in='gei', coord_out=coord_system.lower())
    newname = 'fac_mat_pos_tmp_' + coord_system.lower()
    pos_data = get_data(newname)
    del_data(newname)

    return pos_data.y


def get_z_axis(mag_data):
    """Normalized magnetic field vector"""
    return tnormalize(mag_data.y, return_data=True)


def build_fac_axes(z, refv, mode='yxz'):
    """Construct FAC axes based on defined mode"""
    if mode == 'yxz':
        y = tcrossp(z, refv, return_data=True)
        y = normalize_vectors(y)
        x = tcrossp(y, z, return_data=True)
        return x, y, z
    elif mode == 'xyz':
        x = tcrossp(refv, z, return_data=True)
        x = normalize_vectors(x)
        y = tcrossp(z, x, return_data=True)
        y = normalize_vectors(y)
        return x, y, z
    else:
        logging.error("Bad cross order: " + mode)
        return None, None, None


def get_ref_xgse(N):
    x_axis = np.zeros((N, 3))
    x_axis[:, 0] = 1
    return x_axis


def get_ref_ygsm(pos_data, mag):
    magdat = get_data(mag)
    y_axis = np.zeros((len(magdat.times), 3))
    y_axis[:, 1] = 1
    mag_coord = get_coords(mag).lower()
    arr = cotrans(data_in=y_axis, time_in = magdat.times, coord_in='gsm', coord_out=mag_coord)
    return arr


def get_ref_rgeo(pos_data, mag_var_name, sign=1):
    arr = normalize_vectors(pos_data.y * sign)
    arr = pos_cotrans(pos_data.times, arr, mag_var_name)
    return arr


def get_ref_phi(pos_data, mag_var_name, sign=1):
    polar = xyz_to_polar(pos_data.y)
    phi_deg = polar[:, 2]
    arr = np.empty_like(pos_data.y)
    arr[:, 0] = -np.sin(np.radians(phi_deg)) * sign
    arr[:, 1] = np.cos(np.radians(phi_deg)) * sign
    arr[:, 2] = 0.0
    arr = pos_cotrans(pos_data.times, arr, mag_var_name)
    return arr

def get_ref_phi_sm(pos_data, mag_var_name, sign=1):
    polar = xyz_to_polar(pos_data.y)
    phi_deg = polar[:, 2]
    arr = np.empty_like(pos_data.y)
    arr[:, 0] = -np.sin(np.radians(phi_deg)) * sign
    arr[:, 1] = np.cos(np.radians(phi_deg)) * sign
    arr[:, 2] = 0.0
    arr_gei = cotrans(data_in=arr, time_in=pos_data.times, coord_in='sm', coord_out='gei')
    arr = pos_cotrans(pos_data.times, arr_gei, mag_var_name)
    return arr

# === Dictionary Mapping Option Strings to Functions ===
COORD_FUNCTIONS = {
    "xgse": {"mode": "yxz", "ref": lambda mag, pos: get_ref_xgse(len(get_data(mag).times))},
    "rgeo": {"mode": "yxz", "ref": lambda mag, pos: get_ref_rgeo(pos, mag, 1)},
    "mrgeo": {"mode": "yxz", "ref": lambda mag, pos: get_ref_rgeo(pos, mag, -1)},
    "phigeo": {"mode": "xyz", "ref": lambda mag, pos: get_ref_phi(pos, mag,  1)},
    "mphigeo": {"mode": "xyz", "ref": lambda mag, pos: get_ref_phi(pos, mag, -1)},
    "phism": {"mode": "xyz", "ref": lambda mag, pos: get_ref_phi_sm(pos, mag,  1)},
    "mphism": {"mode": "xyz", "ref": lambda mag, pos: get_ref_phi_sm(pos, mag, -1)},
    "ygsm": {"mode": "xyz", "ref": lambda mag, pos: get_ref_ygsm(pos, mag)}
}

# === Main Interface ===
def fac_matrix_make(mag_var_name, other_dim='Xgse', pos_var_name=None, newname=None):
    """
    Generate a field-aligned coordinate (FAC) transformation matrix from the given magnetic
    field B and (if required) position data, and store it in a tplot variable.

    Parameters
    ----------
    mag_var_name : str
        tplot variable containing the magnetic field data.
    other_dim : str, optional
        The coordinate system option. Must be one of:
        ['xgse','rgeo','phigeo', 'mrgeo','mphigeo','phism','mphism','ygsm','zdsl'].
        (Default is "xgse".)
    pos_var_name : str, optional
        tplot variable containing the spacecraft position data.
        Required for all options except "xgse" (and "zdsl", in development).
    newname : str, optional
        Name of the output tplot variable. Defaults to mag_var_name + "_fac_mat".

    Returns
    -------
    str or None
        The name of the output tplot variable containing the FAC matrix,
        or None if an error occurs.
    """

    mag_data = validate_vector_data(mag_var_name)

    if mag_data is None:
        logging.error('Error reading tplot variable: %s', mag_var_name)
        return

    if newname is None:
        newname = mag_var_name + '_fac_mat'

    # Standardize coordinate option.
    coord_option = other_dim.lower()

    if coord_option not in COORD_FUNCTIONS:
        logging.error("Unsupported transform %s", other_dim)
        return None

    # For certain coordinate systems, a position variable is required.
    required_pos = ["rgeo", "mrgeo", "phigeo", "mphigeo", "phism", "mphism"]
    if coord_option in required_pos and pos_var_name is None:
        logging.error("Need pos_var_name for %s", other_dim)
        return None

    if coord_option in required_pos:
        # All routines assume the position data is originally in GEI.
        if get_coords(pos_var_name).lower() != 'gei':
            logging.error("Position must be in GEI coordinates for tplot variable %s.", pos_var_name)
            return None

        if coord_option in ["phism", "mphism"]:
            # Convert position to sm
            cotrans(name_in=pos_var_name, name_out=pos_var_name + '-sm', coord_out='sm')
            pos_var_name = pos_var_name + '-sm'

        # Interpolate position data onto the magnetic field's time grid.
        interp_pos_data = interpolate_position(pos_var_name, mag_var_name)
        if interp_pos_data is None:
            return None
    else:
        interp_pos_data = None

    # Retrieve the normalized magnetic field vector (FAC Z-axis).
    fac_z_axis = get_z_axis(mag_data)
    if fac_z_axis is None:
        return None

    # Get the reference vector using the transform's reference function.
    ref_vector = COORD_FUNCTIONS[coord_option]["ref"](mag_var_name, interp_pos_data)
    if ref_vector is None:
        return None

    # Build the FAC axes (x, y, z) using the normalized magnetic field and the reference vector.
    x_axis, y_axis, z_axis = build_fac_axes(fac_z_axis, ref_vector, mode=COORD_FUNCTIONS[coord_option]["mode"])
    if x_axis is None:
        return None

    # Construct the FAC matrix.
    num_points = len(mag_data.times)
    fac_mat = np.zeros((num_points, 3, 3))
    fac_mat[:, 0, :] = x_axis
    fac_mat[:, 1, :] = y_axis
    fac_mat[:, 2, :] = z_axis

    # Store the computed FAC matrix into a new tplot variable.
    store_data(newname, data={'x': mag_data.times, 'y': fac_mat})
    mag_data_coords = get_coords(mag_var_name)
    out_coords = "FAC-" + coord_option.upper()
    rotmat_set_coords(newname, mag_data_coords.upper(), out_coords)

    return newname



"""
Unit Tests for minvar function.
"""

import os
import unittest
import numpy as np
from numpy.testing import assert_allclose
import pyspedas
from pyspedas import (
    tplot_restore,
    tvector_rotate,
    data_exists,
    get_data,
    tplot_copy,
    del_data,
    minvar,
    minvar_matrix_make,
    rotmat_get_coords,
    get_coords,
)
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
cotrans_dir = "cotrans_tools"
save_dir = os.path.join(output_dir, cotrans_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


class TestMinvar(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        #  Test tolerance
        cls.tol = 1e-10

        # Define a random data array
        rng = np.random.default_rng(seed=31415)
        cls.rdata = rng.random((10, 3))

        # The SPEDAS script that creates the file: general/tools/python_validate/minvar_python_validate.pro

        # Download tplot files with validation data
        filename = test_data_download_file(
            validation_dir, cotrans_dir, "mva_python_validate.tplot", save_dir
        )

        # Load validation variables from the test file
        del_data("*")
        tplot_restore(filename)

        cls.fgm = get_data("thb_fgs_gsm_mvaclipped1")
        cls.mat = get_data("thb_fgs_gsm_mvaclipped1_mva_mat")
        cls.vals = get_data("mva_vals")
        cls.min = get_data("mva_min")
        cls.int = get_data("mva_int")
        cls.max = get_data("mva_max")
        cls.rot = get_data("thb_fgs_gsm_mvaclipped1_rot")
        tplot_copy("thb_fgs_gsm_mvaclipped1", "idl_thb_fgs_gsm_mvaclipped1")
        tplot_copy(
            "thb_fgs_gsm_mvaclipped1_mva_mat", "idl_thb_fgs_gsm_mvaclipped1_mva_mat"
        )
        tplot_copy("mva_min", "idl_mva_min")
        tplot_copy("mva_int", "idl_mva_int")
        tplot_copy("mva_max", "idl_mva_max")
        tplot_copy("mva_vals", "idl_mva_vals")
        tplot_copy("thb_fgs_gsm_mvaclipped1_rot", "idl_thb_fgs_gsm_mvaclipped1_rot")

        cls.fgm2 = get_data("mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar")
        cls.mat2 = get_data("mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat")
        cls.inp2 = get_data("mms1_fgm_b_gsm_srvy_l2_bvec_bp")
        cls.vals2 = get_data("mva_vals2")
        cls.min2 = get_data("mva_min2")
        cls.int2 = get_data("mva_int2")
        cls.max2 = get_data("mva_max2")
        cls.rot2 = get_data("mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot")
        tplot_copy(
            "mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar",
            "idl_mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar",
        )
        tplot_copy(
            "mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat",
            "idl_mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat",
        )
        tplot_copy("mva_min2", "idl_mva_min2")
        tplot_copy("mva_int2", "idl_mva_int2")
        tplot_copy("mva_max2", "idl_mva_max2")
        tplot_copy("mva_vals2", "idl_mva_vals2")
        tplot_copy(
            "mms1_fgm_b_gsm_srvy_l2_bvec_bp", "idl_mms1_fgm_b_gsm_srvy_l2_bvec_bp"
        )
        tplot_copy(
            "mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot",
            "idl_mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot",
        )

        del_data("thb_*")
        del_data("mms1_*")
        del_data("mva*")

    def test_minvar_basic(self):
        """Test of basic input and output"""

        # Synthetic data of zeros
        data = np.zeros([2, 3])
        vrot, v, w = minvar(data)
        self.assertTrue(np.sum(vrot - data) < self.tol)
        self.assertTrue(np.sum(v - np.diag(np.ones(3))) < self.tol)
        self.assertTrue(np.sum(w - np.zeros(3)) < self.tol)

    def test_minvar_rotation(self):
        """Test of the rotation matrix"""
        vrot, v, w = minvar(self.rdata)
        # Determinant of rotation matrix should be = 1
        self.assertTrue((np.linalg.det(v) - 1) < self.tol)

    def test_minvar_total(self):
        """Test of same square root of total of squares"""
        vrot, v, w = minvar(self.rdata)
        total1 = (self.rdata**2).sum(axis=1)
        total2 = (vrot**2).sum(axis=1)
        # Compare totals
        self.assertTrue(np.sum(total1 - total2) < self.tol)

    def test_minvar_code_coverage(self):
        """Test to cover the code from IDL"""
        data = np.array([[0, 0, 1], [0, 0, 1]])
        vrot, v, w = minvar(data)
        # Case of np.sum(w) == 0.0
        self.assertTrue(w.sum() < self.tol)

        # This should be not Right Handed (?...)
        data = np.array([[0, -1, 1], [-1, -1, 1]])
        # case if YcrossZdotX < 0
        vrot, v, w = minvar(data)
        self.assertTrue(len(vrot) > 0)
        YcrossZdotX = v[0, 0] * (v[1, 1] * v[2, 2] - v[2, 1] * v[1, 2])
        # YcrossZdotX Should be positive after that
        self.assertTrue(YcrossZdotX > 0)

        # should tigger case if v[2, 2] < 0: (?...)
        data = np.array([[-0.1, -0.9, 0.5], [-1, 1, -0.9]])
        vrot, v, w = minvar(data)
        # v[2,2] Should be positive after that
        self.assertTrue(v[2, 2] > 0)

    def test_minvar_matrix_make_day(self):
        trange = ["2007-07-10", "2007-07-11"]
        pyspedas.projects.themis.fgm(probe="c", trange=trange, level="l2", coord="gse")
        minvar_matrix_make(
            "thc_fgs_gse", tstart="2007-07-10/07:54:00", tstop="2007-07-10/07:56:30"
        )
        self.assertTrue(data_exists("thc_fgs_gse_mva_mat"))
        mat_coord = rotmat_get_coords("thc_fgs_gse_mva_mat")
        pyspedas.tvector_rotate(
            "thc_fgs_gse_mva_mat", "thc_fgs_gse", newname="mva_data_day"
        )
        self.assertTrue((data_exists("mva_data_day")))
        out_coord = get_coords("mva_data_day")
        self.assertTrue(len(mat_coord) == 2)
        self.assertEqual(mat_coord[0].lower(), "gse")
        self.assertEqual(mat_coord[1].lower(), "mva")
        self.assertEqual(out_coord.lower(), "mva")

    def test_minvar_matrix_extra_outputs(self):
        trange = ["2007-07-10", "2007-07-11"]
        pyspedas.projects.themis.fgm(probe="c", trange=trange, level="l2", coord="gse")
        minvar_matrix_make(
            "thc_fgs_gse",
            tstart="2007-07-10/07:54:00",
            tstop="2007-07-10/07:56:30",
            evname="mva_lambdas",
            tminname="mva_min",
            tmidname="mva_mid",
            tmaxname="mva_max",
        )
        self.assertTrue(data_exists("thc_fgs_gse_mva_mat"))
        pyspedas.tvector_rotate(
            "thc_fgs_gse_mva_mat", "thc_fgs_gse", newname="mva_data_day"
        )
        self.assertTrue((data_exists("mva_data_day")))
        self.assertTrue(data_exists("mva_lambdas"))
        self.assertTrue(data_exists("mva_min"))
        self.assertTrue(data_exists("mva_min"))
        self.assertTrue(data_exists("mva_max"))

    def test_minvar_matrix_make_hour(self):
        trange = ["2007-07-10", "2007-07-11"]
        pyspedas.projects.themis.fgm(probe="c", trange=trange, level="l2", coord="gse")
        minvar_matrix_make("thc_fgs_gse", twindow=3600, tslide=300)
        self.assertTrue(data_exists("thc_fgs_gse_mva_mat"))
        pyspedas.tvector_rotate(
            "thc_fgs_gse_mva_mat", "thc_fgs_gse", newname="mva_data_hour"
        )
        self.assertTrue((data_exists("mva_data_hour")))

    def test_minvar_matrix_make_hour_interp(self):
        trange = ["2007-07-10", "2007-07-11"]
        pyspedas.projects.themis.fgm(probe="c", trange=trange, level="l2", coord="gse")
        minvar_matrix_make("thc_fgs_gse", twindow=3600, tslide=300)
        self.assertTrue(data_exists("thc_fgs_gse_mva_mat"))
        # Check rotation matrix input and output coords
        in_coord, out_coord = pyspedas.rotmat_get_coords("thc_fgs_gse_mva_mat")
        self.assertEqual(in_coord.upper(), "GSE")
        self.assertEqual(out_coord.upper(), "MVA")
        pyspedas.tvector_rotate(
            "thc_fgs_gse_mva_mat", "thc_fgl_gse", newname="mva_data_hour_interp"
        )
        self.assertTrue((data_exists("mva_data_hour_interp")))
        # Check output vector coordinates
        vec_out_coord = pyspedas.get_coords("mva_data_hour_interp")
        self.assertEqual(vec_out_coord.upper(), "MVA")

    def test_minvar_matrix_make_idl_single(self):
        """Test a minvar_matrix_make call with only a single window"""
        minvar_matrix_make(
            "idl_thb_fgs_gsm_mvaclipped1",
            newname="thb_fgs_gsm_mvaclipped1_mva_mat",
            evname="mva_vals",
            tminname="mva_min",
            tmidname="mva_int",
            tmaxname="mva_max",
        )

        tvector_rotate(
            "thb_fgs_gsm_mvaclipped1_mva_mat",
            "idl_thb_fgs_gsm_mvaclipped1",
            newname="thb_fgs_gsm_mvaclipped1_rot",
        )
        vals = get_data("mva_vals")
        assert_allclose(vals.y, self.vals.y, rtol=1e-05, atol=1e-03)
        assert_allclose(vals.times, self.vals.times, atol=1e-06)
        mva_mat = get_data("thb_fgs_gsm_mvaclipped1_mva_mat")
        assert_allclose(mva_mat.y, self.mat.y, rtol=1e-05, atol=1e-03)
        mva_min = get_data("mva_min")
        assert_allclose(mva_min.y, self.min.y, rtol=1e-05, atol=1e-03)
        mva_int = get_data("mva_int")
        assert_allclose(mva_int.y, self.int.y, rtol=1e-05, atol=1e-03)
        mva_max = get_data("mva_max")
        assert_allclose(mva_max.y, self.max.y, rtol=1e-05, atol=1e-03)
        mva_rot = get_data("thb_fgs_gsm_mvaclipped1_rot")
        assert_allclose(mva_rot.y, self.rot.y, rtol=1e-05, atol=1e-03)

    def test_minvar_matrix_make_idl_multi(self):
        """Test a minvar_matrix_make call that returns multiple matrices"""
        minvar_matrix_make(
            "idl_mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar",
            newname="mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat",
            twindow=16.0,
            tslide=4.0,
            evname="mva_vals2",
            tminname="mva_min2",
            tmidname="mva_int2",
            tmaxname="mva_max2",
        )
        tvector_rotate(
            "mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat",
            "idl_mms1_fgm_b_gsm_srvy_l2_bvec_bp",
            newname="mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot",
        )
        vals = get_data("mva_vals2")
        assert_allclose(vals.times, self.vals2.times, atol=1e-06)
        assert_allclose(vals.y, self.vals2.y, rtol=1e-05, atol=1e-06)
        mva_mat = get_data("mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat")
        assert_allclose(mva_mat.y, self.mat2.y, rtol=1e-03, atol=1e-05)
        mva_min = get_data("mva_min2")
        assert_allclose(mva_min.y, self.min2.y, rtol=1e-03, atol=1e-06)
        mva_int = get_data("mva_int2")
        assert_allclose(mva_int.y, self.int2.y, rtol=1e-03, atol=1e-06)
        mva_max = get_data("mva_max2")
        assert_allclose(mva_max.y, self.max2.y, rtol=1e-03, atol=1e-06)
        mva_rot = get_data("mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot")
        assert_allclose(mva_rot.y, self.rot2.y, rtol=1e-05, atol=1e-06)


if __name__ == "__main__":
    unittest.main()



"""
Transform strings to datetime decimals.
"""
from dateutil import parser
from datetime import datetime, timezone
import numpy as np
from collections.abc import Iterable


def time_float_one(s_time=None):
    """
    Transform one datetime from string to decimal.

    Parameters
    ----------
    s_time : str, optional
        Input string representing a datetime. If not provided, the current time is used.

    Returns
    -------
    float_time : float
        The decimal representation of the input datetime, or the current local datetime if no input is provided.

    Examples
    --------
    >>> from pyspedas import time_float_one
    >>> time_float_one('2023-03-25 12:00:00')
    1679745600.0

    """
    if s_time is None:
        s_time = str(datetime.now())

    if isinstance(s_time, (int, float, np.integer, np.float64)):
        return float(s_time)

    try:
        in_datetime = parser.isoparse(s_time)
    except ValueError:
        in_datetime = parser.parse(s_time)

    float_time = in_datetime.replace(tzinfo=timezone.utc).timestamp()

    return float_time


def time_float(str_time=None):
    """
    Transform a list of datetimes from string to decimal.

    Parameters
    ----------
    str_time : str or list of str, optional
        The input datetime(s) as a string or a list of strings. If not provided the current time is used.

    Returns
    -------
    time_list : list of float or float
        The decimal representation of the input datetime(s) as a list of floats. If the input is a single
        datetime string, the list will contain only one element. The current local datetime is used if no input is provided.

    Examples
    --------
    >>> from pyspedas import time_float
    >>> time_float(['2023-03-25 12:00:00', '2023-03-26 12:00:00'])
    [1679745600.0, 1679832000.0]

    >>> time_float('2023-03-25 12:00:00')
    1679745600.0

    >>> time_float()
    # Returns the current timestamp as a list containing one float.

    Note
    ----
    This function is to time_double.pro in IDL SPEDAS.
    """

    if str_time is None:
        return time_float_one()

    if isinstance(str_time, str):
        return time_float_one(str_time)

    time_list = list()
    if isinstance(str_time, Iterable):
        for t in str_time:
            time_list.append(time_float_one(t))
        return time_list
    else:
        return time_float_one(str_time)


def time_double(str_time=None):
    """
    Transform a list of datetimes from string to decimal.

    Parameters
    ----------
    str_time : str or list of str, optional
        The input datetime(s) as a string or a list of strings. If not provided the current time is used.

    Returns
    -------
    time_list : list of float or float
        The decimal representation of the input datetime(s) as a list of floats. If the input is a single
        datetime string, the list will contain only one element. The current local datetime is used if no input is provided.

    Examples
    --------
    >>> from pyspedas import time_double
    >>> time_double(['2023-03-25 12:00:00', '2023-03-26 12:00:00'])
    [1679745600.0, 1679832000.0]

    >>> time_double('2023-03-25 12:00:00')
    1679745600.0

    >>> time_double()
    # Returns the current timestamp as a list containing one float.

    Note
    ----
    This function is an alilas of time_float. This function is to time_double.pro in IDL SPEDAS.

    """
    return time_float(str_time)
